"""
This module provides a collection of time handing functions.
"""
import re
import textwrap
import contextlib
from datetime import date, datetime
from functools import singledispatch

import numpy as np

import astropy.table
import astropy.time
import astropy.units as u
from astropy.time import Time, TimeDelta
from astropy.time import conf as astropy_time_conf
from astropy.utils.introspection import minversion

from sunpy import log
# This is not called but imported to register time formats
from sunpy.time.timeformats import *  # NOQA
from sunpy.util.decorators import add_common_docstring

__all__ = [
    'find_time', 'parse_time', 'is_time',
    'is_time_in_given_format', 'is_time_equal',
    'julian_centuries'
]

# Mapping of time format codes to regular expressions.
REGEX = {
    '.': r'\.',
    '%Y': r'(?P<year>\d{4})',
    '%j': r'(?P<dayofyear>\d{3})',
    '%m': r'(?P<month>\d{1,2})',
    '%d': r'(?P<day>\d{1,2})',
    '%H': r'(?P<hour>\d{1,2})',
    '%M': r'(?P<minute>\d{1,2})',
    '%S': r'(?P<second>\d{1,2})',
    '%f': r'(?P<microsecond>\d+)',
    '%b': r'(?P<month_str>[a-zA-Z]+)',
}
# DO NOT SORT THIS LIST.
# The string parsing is dependent on the specific order within this list.
TIME_FORMAT_LIST = [
    "%Y-%m-%dT%H:%M:%S.%f",  # Example 2007-05-04T21:08:12.999999
    "%Y/%m/%dT%H:%M:%S.%f",  # Example 2007/05/04T21:08:12.999999
    "%Y-%m-%dT%H:%M:%S.%fZ",  # Example 2007-05-04T21:08:12.999Z
    "%Y-%m-%dT%H:%M:%S",  # Example 2007-05-04T21:08:12
    "%Y/%m/%dT%H:%M:%S",  # Example 2007/05/04T21:08:12
    "%Y-%m-%dT%H:%M:%SZ",  # Example 2007-05-04T21:08:12Z
    "%Y-%m-%dT%H%M%SZ",  # Example 2010-08-30T111402Z
    "%Y-%m-%dT%H%M%S",  # Example 2010-08-30T111402
    "%Y%m%dT%H%M%S.%f",  # Example 20070504T210812.999999
    "%Y%m%dT%H%M",  # Example 20070504T2108 , Should precede "%Y%m%dT%H%M%S".
    "%Y%m%dT%H%M%S",  # Example 20070504T210812
    "%Y/%m/%d %H:%M:%S",  # Example 2007/05/04 21:08:12
    "%Y/%m/%d %H:%M",  # Example 2007/05/04 21:08
    "%Y/%m/%d %H:%M:%S.%f",  # Example 2007/05/04 21:08:12.999999
    "%Y-%m-%d %H:%M:%S.%f",  # Example 2007-05-04 21:08:12.999999
    "%Y-%m-%d %H:%M:%S",  # Example 2007-05-04 21:08:12
    "%Y-%m-%d %H:%M",  # Example 2007-05-04 21:08
    "%Y-%b-%d %H:%M:%S.%f",  # Example 2007-May-04 21:08:12.999999
    "%Y-%b-%d %H:%M:%S",  # Example 2007-May-04 21:08:12
    "%Y-%b-%d %H:%M",  # Example 2007-May-04 21:08
    "%Y-%b-%d",  # Example 2007-May-04
    "%Y-%m-%d",  # Example 2007-05-04
    "%Y/%m/%d",  # Example 2007/05/04
    "%d-%b-%Y",  # Example 04-May-2007
    "%d-%b-%Y %H:%M:%S",  # Example 04-May-2007 21:08:12
    "%d-%b-%Y %H:%M:%S.%f",  # Example 04-May-2007 21:08:12.999999
    "%Y%m%d_%H%M",  # Example 20070504_2108 , Should precede "%Y%m%d_%H%M%S".
    "%Y%m%d_%H%M%S",  # Example 20070504_210812
    "%Y:%j:%H:%M:%S",  # Example 2012:124:21:08:12
    "%Y:%j:%H:%M:%S.%f",  # Example 2012:124:21:08:12.999999
    "%Y%m%d%H%M" ,   # Example 201401041205 , Should precede "%Y%m%d%H%M%S".
    "%Y%m%d%H%M%S",  # Example 20140101000001 (JSOC/VSO Export/Downloads)
    "%Y.%m.%d_%H:%M:%S_TAI",  # Example 2016.05.04_21:08:12_TAI - JSOC
    "%Y.%m.%d_%H:%M:%S.%f_TAI",  # Example 2019.09.15_00:00:02.898_TAI - JSOC
    "%Y.%m.%d_%H:%M:%S_UTC",  # Example 2016.05.04_21:08:12_UTC - JSOC
    "%Y.%m.%d_%H:%M:%S",  # Example 2016.05.04_21:08:12 - JSOC
    "%Y/%m/%dT%H:%M",  # Example 2007/05/04T21:08
]

_ONE_DAY_TIMEDELTA = TimeDelta(1 * u.day)


def is_time_equal(t1, t2):
    """
    Work around for https://github.com/astropy/astropy/issues/6970.

    Remove the usage of this function once the fix is in place.
    """
    if abs(t1 - t2) < 1 * u.nanosecond:
        return True
    return False


def _group_or_none(match, group, fun):
    try:
        ret = match.group(group)
    except IndexError:
        return None
    else:
        return fun(ret)


def _n_or_eq(a, b):
    return a is None or a == b


def _regex_parse_time(inp, format):
    # Parser for finding out the minute value so we can adjust the string
    # from 24:00:00 to 00:00:00 the next day because strptime does not
    # understand the former.
    for key, value in REGEX.items():
        format = format.replace(key, value)
    match = re.match(f"{format}$", inp)
    if match is None:
        return None, None

    found_groups = match.groupdict()

    # Special handling to strip any excess zeros beyond the six digits of the microsecond field
    if "microsecond" in found_groups and re.match(r"\d{6}0+", match.group("microsecond")):
        from_, to = match.span("microsecond")
        inp = inp[:from_] + match.group("microsecond")[:6] + inp[to:]
        match = re.match(format, inp)

    # Special handling to add a day if the hour is 24 and the minute/second/microsecond are all zero
    add_one_day = False
    if "hour" in found_groups and match.group("hour") == "24":
        if not all(
                _n_or_eq(_group_or_none(match, g, int), 00)
                for g in ["minute", "second", "microsecond"]):
            raise ValueError
        from_, to = match.span("hour")
        inp = inp[:from_] + "00" + inp[to:]
        add_one_day = True
        match = re.match(format, inp)

    return inp, add_one_day


def find_time(string, format):
    """
    Return iterator of occurrences of date formatted with format in string.

    Currently supported format codes:
    """
    re_format = format
    for key, value in REGEX.items():
        re_format = re_format.replace(key, value)
    matches = re.finditer(re_format, string)
    for match in matches:
        try:
            matchstr = string[slice(*match.span())]
            dt = datetime.strptime(matchstr, format)
        except ValueError:
            continue
        else:
            yield dt


find_time.__doc__ += ', '.join(list(REGEX.keys()))


@singledispatch
def convert_time(time_string, format=None, **kwargs):
    # default case when no type matches
    return Time(time_string, format=format, **kwargs)


# Only register pandas if we can import pandas
try:
    import pandas

    @convert_time.register(pandas.Timestamp)
    def convert_time_pandasTimestamp(time_string, **kwargs):
        return Time(time_string.asm8)

    @convert_time.register(pandas.Series)
    @convert_time.register(pandas.Index)
    def convert_time_pandasSeries(time_string, **kwargs):
        return convert_time(time_string.tolist(), **kwargs)

    @convert_time.register(pandas.DatetimeIndex)
    def convert_time_pandasDatetimeIndex(time_string, **kwargs):
        return Time(time_string.tolist(), **kwargs)

except ImportError:
    pass


@convert_time.register(datetime)
def convert_time_datetime(time_string, **kwargs):
    return Time(time_string, **kwargs)


@convert_time.register(date)
def convert_time_date(time_string, **kwargs):
    return Time(time_string.isoformat(), **kwargs)


@convert_time.register(tuple)
def convert_time_tuple(time_string, **kwargs):
    # Make sure there are enough values to unpack
    time_string = (time_string + (0, ) * 7)[:7]
    return Time('{}-{}-{}T{}:{}:{}.{:06}'.format(*time_string), **kwargs)


@convert_time.register(np.datetime64)
def convert_time_npdatetime64(time_string, **kwargs):
    return Time(str(time_string.astype('M8[ns]')), **kwargs)


@convert_time.register(np.ndarray)
def convert_time_npndarray(time_string, **kwargs):
    if 'datetime64' in str(time_string.dtype):
        return Time([str(dt.astype('M8[ns]')) for dt in time_string], **kwargs)
    else:
        return convert_time.dispatch(object)(time_string, **kwargs)


@convert_time.register(astropy.time.Time)
def convert_time_astropy(time_string, **kwargs):
    return time_string


@convert_time.register(astropy.table.Column)
def convert_time_astropy_table_column(time_string, **kwargs):
    return convert_time(time_string.tolist(), **kwargs)


@convert_time.register(str)
@convert_time.register(list)
def convert_time_str(time_string, **kwargs):
    is_single_string = isinstance(time_string, str)
    is_string_list = isinstance(time_string, list) and isinstance(time_string[0], str)

    if is_single_string or is_string_list:
        first_item = time_string[0] if is_string_list else time_string
        if 'TAI' in first_item:
            kwargs['scale'] = 'tai'

        for time_format in TIME_FORMAT_LIST:
            try:
                try:
                    ts, add_one_day = _regex_parse_time(first_item, time_format)
                except TypeError:
                    break
                if ts is None:
                    continue
                if is_single_string:
                    t = Time.strptime(ts, time_format, **kwargs)
                    if add_one_day:
                        t += _ONE_DAY_TIMEDELTA
                else:
                    # For a list of strings, we do not try to correct 24:00:00
                    t = Time.strptime(time_string, time_format, **kwargs)
                return t
            except ValueError:
                pass

        log.debug("No matching sunpy format found for %s, so falling back to astropy formats", first_item)

    # If the string format does not match one of ours, we need to protect against a bad interaction
    # between astropy's C fast parser and numpy>=2.3
    # https://github.com/astropy/astropy/issues/18254
    # TODO: once the bug is fixed, skip this protection for fixed versions of astropy/numpy
    if is_string_list and len(time_string) > 500 and astropy_time_conf.use_fast_parser == "True" and minversion(np, "2.3.0"):
        log.debug("Disabling astropy's fast time parsing due to a known bug with long lists")
        ctx = astropy_time_conf.set_temp("use_fast_parser", "False")
    else:
        ctx = contextlib.nullcontext()

    with ctx:
        # when no format matches, call default function
        return convert_time.dispatch(object)(time_string, **kwargs)


def _variables_for_parse_time_docstring():
    ret = {}

    example_time = datetime(2017, 1, 1, 11, 10, 9)
    example_parse_time = [example_time.strftime(f) for f in TIME_FORMAT_LIST]
    example_parse_time = "\n      ".join(example_parse_time)
    ret['parse_time_formats'] = example_parse_time

    types = list(convert_time.registry.keys())
    types.remove(object)
    # Do Builtins
    types2 = [t.__qualname__ for t in types if t.__module__ == "builtins"]
    # # Do all the non-special ones where we take the package name and the class
    types2 += [t.__module__.split(".")[0] + "." +
               t.__qualname__ for t in types if not t.__module__.startswith(("builtins", "astropy"))]
    # Special case astropy.time where we need the subpackage
    types2 += ["astropy.time." +
               t.__qualname__ for t in types if t.__module__.startswith("astropy.time")]
    parse_time_types = str(types2)[1:-1].replace("'", "`")
    ret['parse_time_types'] = parse_time_types
    ret['parse_time_desc'] = """
                             Any time input, will be passed into `~sunpy.time.parse_time`.
                             """
    try:
        # Need to try importing cdflib, as if it is present it will register
        # extra formats with time
        import cdflib  # NOQA
    except Exception:
        pass
    ret['astropy_time_formats'] = textwrap.fill(str(list(astropy.time.Time.FORMATS.keys())),
                                                subsequent_indent=' '*10)

    return ret


@add_common_docstring(**_variables_for_parse_time_docstring())
def parse_time(time_string, *, format=None, **kwargs):
    """
    Takes a time input and will parse and return a `astropy.time.Time`.

    Parameters
    ----------
    time_string : {parse_time_types}
        Time to parse.

    format : `str`, optional
        Specifies the format user has provided the time_string in.
        We support the same formats as `astropy.time.Time`, which are::

          >>> list(astropy.time.Time.FORMATS)
          {astropy_time_formats}

    **kwargs :
        Additional keyword arguments are passed to `astropy.time.Time`

    Returns
    -------
    `astropy.time.Time`
        `~astropy.time.Time` corresponding to input time string.

    Examples
    --------
    >>> import sunpy.time
    >>> sunpy.time.parse_time('2012/08/01')
    <Time object: scale='utc' format='isot' value=2012-08-01T00:00:00.000>
    >>> sunpy.time.parse_time('2016.05.04_21:08:12_TAI')
    <Time object: scale='tai' format='isot' value=2016-05-04T21:08:12.000>

    Notes
    -----
    Additional keyword arguments can be passed to `astropy.time.Time`

    The list of time formats are show by the following examples::

      {parse_time_formats}
    """
    if isinstance(time_string, str) and time_string == 'now':
        rt = Time.now()
    else:
        rt = convert_time(time_string, format=format, **kwargs)

    return rt


def is_time(time_string, time_format=None):
    """
    Returns true if the input is a valid date/time representation.

    Parameters
    ----------
    time_string : `int`, `float`, ``time_string``, `datetime`
        Date to parse.
    time_format : basestring, utime, `datetime`, optional
        Specifies the format user has provided the ``time_string`` in.

    Returns
    -------
    `bool`
        `True` if can be parsed by `sunpy.time.parse_time`.

    Examples
    --------
    >>> import sunpy.time
    >>> sunpy.time.is_time('2012/08/01')
    True
    """
    if time_string is None:
        return False
    elif isinstance(time_string, Time):
        return True

    try:
        parse_time(time_string, format=time_format)
    except ValueError:
        return False
    else:
        return True


def is_time_in_given_format(time_string, time_format):
    """
    Tests whether a time string is formatted according to the given time
    format.

    Parameters
    ----------
    time_string : `str`
        Date to parse.
    time_format : basestring, utime, `datetime`, optional
        Specifies the format user has provided the ``time_string`` in.

    Returns
    -------
    `bool`
        `True` if can it if follows the ``time_format``.
    """
    try:
        datetime.strptime(time_string, time_format)
        return True
    except ValueError:
        return False


def julian_centuries(t='now'):
    """
    Returns the number of Julian centuries since J1900.0 (noon on 1900 January
    0).
    """
    DAYS_IN_JULIAN_CENTURY = 36525.0

    # J1900.0 is 2415021.0
    return (parse_time(t).jd - 2415020.0) / DAYS_IN_JULIAN_CENTURY



"""
Common solar physics coordinate systems.

This submodule implements various solar physics coordinate frames for use with
the `astropy.coordinates` module.
"""
import os
import re
import traceback

import numpy as np

import astropy.units as u
from astropy.constants import R_earth
from astropy.coordinates import Attribute, ConvertError, Latitude, Longitude, QuantityAttribute
from astropy.coordinates.baseframe import BaseCoordinateFrame, RepresentationMapping
from astropy.coordinates.representation import (
    CartesianDifferential,
    CartesianRepresentation,
    CylindricalRepresentation,
    SphericalDifferential,
    SphericalRepresentation,
    UnitSphericalRepresentation,
)
from astropy.time import Time
from astropy.utils.data import download_file

from sunpy import log
from sunpy.sun.constants import radius as _RSUN
from sunpy.time.time import _variables_for_parse_time_docstring
from sunpy.util.decorators import add_common_docstring
from sunpy.util.exceptions import warn_user
from .frameattributes import ObserverCoordinateAttribute, TimeFrameAttributeSunPy

_J2000 = Time('J2000.0', scale='tt')

__all__ = ['SunPyBaseCoordinateFrame', 'BaseHeliographic', 'BaseMagnetic',
           'HeliographicStonyhurst', 'HeliographicCarrington',
           'Heliocentric', 'Helioprojective', 'HelioprojectiveRadial',
           'HeliocentricEarthEcliptic', 'GeocentricSolarEcliptic',
           'HeliocentricInertial', 'GeocentricEarthEquatorial',
           'Geomagnetic', 'SolarMagnetic', 'GeocentricSolarMagnetospheric']


def _frame_parameters():
    """
    Returns formatting dictionary to use with add_common_docstring to populate frame docstrings
    """
    ret = {}

    ret['data'] = ("data : `~astropy.coordinates.BaseRepresentation` or ``None``\n"
                   "    A representation object or ``None`` to have no data\n"
                   "    (or use the coordinate component arguments, see below).")
    ret['common'] = (f"obstime : {_variables_for_parse_time_docstring()['parse_time_types']}\n"
                     "    The time of the observation. This is used to determine the\n"
                     "    position of solar-system bodies (e.g., the Sun and the Earth) as\n"
                     "    needed to define the origin and orientation of the frame.\n"
                     "representation_type : `~astropy.coordinates.BaseRepresentation`, str, optional\n"
                     "    A representation class or string name of a representation class.\n"
                     "    This may change the valid coordinate component arguments from the\n"
                     "    defaults (see above). For example, passing\n"
                     "    ``representation_type='cartesian'`` will make the frame expect\n"
                     "    Cartesian coordinate component arguments (typically, ``x``, ``y``,\n"
                     "    and ``z``).\n"
                     "copy : bool, optional\n"
                     "    If `True` (default), make copies of the input coordinate arrays.")
    ret['lonlat'] = ("lon : `~astropy.coordinates.Angle` or `~astropy.units.Quantity`, optional\n"
                     "    The longitude coordinate for this object (``lat`` must also be\n"
                     "    given and ``data`` must be ``None``).\n"
                     "    Not needed if ``data`` is given.\n"
                     "lat : `~astropy.coordinates.Angle` or `~astropy.units.Quantity`, optional\n"
                     "    The latitude coordinate for this object (``lon`` must also be\n"
                     "    given and ``data`` must be ``None``).\n"
                     "    Not needed if ``data`` is given.")
    ret['radius'] = ("radius : `~astropy.units.Quantity`, optional\n"
                     "    The radial distance coordinate from Sun center for this object.\n"
                     "    Defaults to the radius of the Sun. Not needed if ``data`` is given.")
    ret['distance_sun'] = ("distance : `~astropy.units.Quantity`, optional\n"
                           "    The distance coordinate from Sun center for this object.\n"
                           "    Not needed if ``data`` is given.")
    ret['distance_earth'] = ("distance : `~astropy.units.Quantity`, optional\n"
                             "    The distance coordinate from Earth center for this object.\n"
                             "    Not needed if ``data`` is given.")
    ret['xyz'] = ("x : `~astropy.units.Quantity`, optional\n"
                  "    X-axis coordinate for this object. Not needed if ``data`` is given.\n"
                  "y : `~astropy.units.Quantity`, optional\n"
                  "    Y-axis coordinate for this object. Not needed if ``data`` is given.\n"
                  "z : `~astropy.units.Quantity`, optional\n"
                  "    Z-axis coordinate for this object. Not needed if ``data`` is given.")
    ret['observer'] = ("observer : `~sunpy.coordinates.frames.HeliographicStonyhurst`, str\n"
                       "    The location of the observer. If a string is provided,\n"
                       "    it must be a solar system body that can be parsed by\n"
                       "    `~sunpy.coordinates.ephemeris.get_body_heliographic_stonyhurst`\n"
                       "    at the time ``obstime``. Defaults to Earth center.")
    ret['rsun'] = ("rsun : `~astropy.units.Quantity`\n"
                   "    The radius of the Sun in length units. Used to convert a 2D\n"
                   "    coordinate (i.e., no ``radius`` component) to a 3D coordinate by\n"
                   "    assuming that the coordinate is on the surface of the Sun. Defaults\n"
                   "    to the photospheric radius as defined in `sunpy.sun.constants`.")
    ret['equinox'] = (f"equinox : {_variables_for_parse_time_docstring()['parse_time_types']}\n"
                      "    The date for the mean vernal equinox.\n"
                      "    Defaults to the J2000.0 equinox.")
    ret['magnetic_model'] = ("magnetic_model : `str`\n"
                             "    The IGRF model to use for determining the orientation of\n"
                             "    Earth's magnetic dipole pole. The supported options are\n"
                             "    ``'igrf13'`` (default), ``'igrf12'``, ``'igrf11'``, and\n"
                             "    ``'igrf10'``.")
    ret['igrf_reference'] = ("* `International Geomagnetic Reference Field (IGRF) "
                             "<https://www.ngdc.noaa.gov/IAGA/vmod/igrf.html>`__")

    return ret


class SunPyBaseCoordinateFrame(BaseCoordinateFrame):
    """
    Base class for sunpy coordinate frames.

    This class is not intended to be used directly and has no transformations defined.

    * Defines the frame attribute ``obstime`` for observation time.
    * Defines a default wrap angle of 180 degrees for longitude in spherical coordinates,
      which can be overridden via the class variable ``_wrap_angle``.
    * Inject a nice way of representing the object which the coordinate represents.
    """
    obstime = TimeFrameAttributeSunPy()

    default_representation = SphericalRepresentation
    default_differential = SphericalDifferential

    frame_specific_representation_info = {
        SphericalDifferential: [RepresentationMapping('d_lon', 'd_lon', u.arcsec/u.s),
                                RepresentationMapping('d_lat', 'd_lat', u.arcsec/u.s),
                                RepresentationMapping('d_distance', 'd_distance', u.km/u.s)],
    }

    _wrap_angle = 180*u.deg  # for longitude in spherical coordinates

    def __init__(self, *args, **kwargs):
        self.object_name = None

        # If wrap_longitude=False is passed in, do not impose a specific wrap angle for the frame
        if not kwargs.pop('wrap_longitude', True):
            self._wrap_angle = None

        # If obstime is not provided but observer has an obstime, use that as the obstime
        if 'obstime' not in kwargs and 'observer' in kwargs and getattr(kwargs['observer'], 'obstime', None) is not None:
            kwargs['obstime'] = kwargs['observer'].obstime

        super().__init__(*args, **kwargs)

        # If obstime is specified, treat the default observer (None) as explicitly set
        if self.obstime is not None and self.is_frame_attr_default('observer'):
            self._attr_names_with_defaults.remove('observer')

        return

    def represent_as(self, base, s='base', in_frame_units=False):
        data = super().represent_as(base, s, in_frame_units=in_frame_units)

        # If a frame wrap angle is set, use that wrap angle for any spherical representations.
        if self._wrap_angle is not None and \
           isinstance(data, UnitSphericalRepresentation | SphericalRepresentation):
            data.lon.wrap_angle = self._wrap_angle
        return data

    def __str__(self):
        # We override this here so that when you print a SkyCoord it shows the
        # observer as the string and not the whole massive coordinate.
        if getattr(self, "object_name", None):
            return f"<{self.__class__.__name__} Coordinate for '{self.object_name}'>"
        else:
            return super().__str__()

    @property
    def _is_2d(self):
        return (self._data is not None and self._data.norm().unit is u.one
                and u.allclose(self._data.norm(), 1*u.one))


class BaseHeliographic(SunPyBaseCoordinateFrame):
    """
    Base class for HeliographicCarrington (HGC) and HeliographicStonyhurst (HGS) frames.

    This class is not intended to be used directly and has no transformations defined.
    """
    frame_specific_representation_info = {
        SphericalRepresentation: [RepresentationMapping('lon', 'lon', u.deg),
                                  RepresentationMapping('lat', 'lat', u.deg),
                                  RepresentationMapping('distance', 'radius', None)],
        SphericalDifferential: [RepresentationMapping('d_lon', 'd_lon', u.arcsec/u.s),
                                RepresentationMapping('d_lat', 'd_lat', u.arcsec/u.s),
                                RepresentationMapping('d_distance', 'd_radius', u.km/u.s)],
    }

    rsun = QuantityAttribute(default=_RSUN, unit=u.km)

    def make_3d(self):
        """
        Returns a fully 3D coordinate based on this coordinate.

        If this coordinate is only 2D (i.e., no ``radius`` component) or is a
        unit vector (i.e., the norm of the coordinate is unity), a new
        coordinate is created that corresponds to the surface of the Sun.
        That is, the 3D coordinate will retain the ``lon`` and ``lat``, and
        ``radius`` will be set to the frame's ``rsun`` frame attribute.

        If this coordinate is already fully 3D, it is directly returned, even
        if it does not lie on the surface of the Sun.

        Returns
        -------
        frame : `~sunpy.coordinates.frames.BaseHeliographic`
            The fully 3D coordinate
        """
        if self._is_2d:
            return self.realize_frame(self._data * self.rsun)

        # The coordinate is already 3D
        return self


@add_common_docstring(**_frame_parameters())
class HeliographicStonyhurst(BaseHeliographic):
    """
    A coordinate or frame in the Stonyhurst Heliographic (HGS) system.

    - The origin is the center of the Sun.
    - The Z-axis (+90 degrees latitude) is aligned with the Sun's north pole.
    - The X-axis (0 degrees longitude and 0 degrees latitude) is aligned with the projection of
      the Sun-Earth line onto the Sun's equatorial plane.

    This system is also know as the Heliocentric Earth Equatorial (HEEQ) system when
    represented using Cartesian components.

    A new instance can be created using the following signatures
    (note that if supplied, ``obstime`` and ``representation_type`` must be
    keyword arguments)::

        HeliographicStonyhurst(lon, lat, obstime=obstime)
        HeliographicStonyhurst(lon, lat, radius, obstime=obstime)
        HeliographicStonyhurst(x, y, z, representation_type='cartesian', obstime=obstime)

    Parameters
    ----------
    {data}
    {lonlat}
    {radius}
    {rsun}
    {common}

    Examples
    --------
    >>> from astropy.coordinates import SkyCoord
    >>> import sunpy.coordinates
    >>> import astropy.units as u
    >>> sc = SkyCoord(1*u.deg, 1*u.deg, 2*u.km,
    ...               frame="heliographic_stonyhurst",
    ...               obstime="2010/01/01T00:00:45")
    >>> sc
    <SkyCoord (HeliographicStonyhurst: obstime=2010-01-01T00:00:45.000, rsun=695700.0 km): (lon, lat, radius) in (deg, deg, km)
        (1., 1., 2.)>
    >>> sc.frame
    <HeliographicStonyhurst Coordinate (obstime=2010-01-01T00:00:45.000, rsun=695700.0 km): (lon, lat, radius) in (deg, deg, km)
        (1., 1., 2.)>
    >>> sc = SkyCoord(HeliographicStonyhurst(-10*u.deg, 2*u.deg))
    >>> sc
    <SkyCoord (HeliographicStonyhurst: obstime=None, rsun=695700.0 km): (lon, lat) in deg
        (-10., 2.)>
    >>> sc = SkyCoord(CartesianRepresentation(0*u.km, 45*u.km, 2*u.km),
    ...               obstime="2011/01/05T00:00:50",
    ...               frame="heliographic_stonyhurst")
    >>> sc
    <SkyCoord (HeliographicStonyhurst: obstime=2011-01-05T00:00:50.000, rsun=695700.0 km): (lon, lat, radius) in (deg, deg, km)
    (90., 2.54480438, 45.04442252)>
    """
    name = "heliographic_stonyhurst"

    def _apply_diffrot(self, duration, rotation_model):
        oldrepr = self.spherical

        from sunpy.sun.models import differential_rotation
        log.debug(f"Applying {duration} of solar rotation")
        newlon = oldrepr.lon + differential_rotation(duration,
                                                     oldrepr.lat,
                                                     model=rotation_model,
                                                     frame_time='sidereal')
        newrepr = SphericalRepresentation(newlon, oldrepr.lat, oldrepr.distance)

        return self.realize_frame(newrepr)


@add_common_docstring(**_frame_parameters())
class HeliographicCarrington(BaseHeliographic):
    """
    A coordinate or frame in the Carrington Heliographic (HGC) system.

    - The origin is the center of the Sun.
    - The Z-axis (+90 degrees latitude) is aligned with the Sun's north pole.
    - The X-axis and Y-axis rotate with a period of 25.38 days.

    This system differs from Stonyhurst Heliographic (HGS) in its definition of longitude. This
    longitude is an "apparent" longitude because it takes into account the time it takes for light
    to travel from the Sun's surface to the observer (see :ref:`sunpy-topic-guide-coordinates-carrington`).
    Thus, the observer needs to be specified to be able to transform to any other coordinate frame.

    A new instance can be created using the following signatures
    (note that if supplied, ``obstime`` and ``observer`` must be a keyword argument)::

        HeliographicCarrington(lon, lat, obstime=obstime, observer=observer)
        HeliographicCarrington(lon, lat, radius, obstime=obstime, observer=observer)

    If you want to define the location in HGC such that the observer for the coordinate frame is
    the same as that location (e.g., the location of an observatory in its corresponding HGC
    frame), use ``observer='self'``::

        HeliographicCarrington(lon, lat, radius, obstime=obstime, observer='self')

    Parameters
    ----------
    {data}
    {lonlat}
    {radius}
    {observer}
    {rsun}
    {common}

    Examples
    --------
    >>> from astropy.coordinates import SkyCoord
    >>> import sunpy.coordinates
    >>> import astropy.units as u
    >>> sc = SkyCoord(1*u.deg, 2*u.deg, 3*u.km,
    ...               frame="heliographic_carrington",
    ...               observer="earth",
    ...               obstime="2010/01/01T00:00:30")
    >>> sc
    <SkyCoord (HeliographicCarrington: obstime=2010-01-01T00:00:30.000, rsun=695700.0 km, observer=<HeliographicStonyhurst Coordinate for 'earth'>): (lon, lat, radius) in (deg, deg, km)
        (1., 2., 3.)>

    >>> sc = SkyCoord([1,2,3]*u.deg, [4,5,6]*u.deg, [5,6,7]*u.km,
    ...               obstime="2010/01/01T00:00:45",
    ...               observer="self",
    ...               frame="heliographic_carrington")
    >>> sc
    <SkyCoord (HeliographicCarrington: obstime=2010-01-01T00:00:45.000, rsun=695700.0 km, observer=self): (lon, lat, radius) in (deg, deg, km)
        [(1., 4., 5.), (2., 5., 6.), (3., 6., 7.)]>

    >>> sc = SkyCoord(CartesianRepresentation(0*u.km, 45*u.km, 2*u.km),
    ...               obstime="2011/01/05T00:00:50",
    ...               frame="heliographic_carrington")
    >>> sc
    <SkyCoord (HeliographicCarrington: obstime=2011-01-05T00:00:50.000, rsun=695700.0 km, observer=None): (lon, lat, radius) in (deg, deg, km)
        (90., 2.54480438, 45.04442252)>
    """
    name = "heliographic_carrington"
    _wrap_angle = 360*u.deg

    observer = ObserverCoordinateAttribute(HeliographicStonyhurst)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not isinstance(self.observer, BaseCoordinateFrame) and self.observer == 'self' and self._is_2d:
            raise ValueError("Full 3D coordinate (including radius) must be specified "
                             "when observer='self'.")


@add_common_docstring(**_frame_parameters())
class Heliocentric(SunPyBaseCoordinateFrame):
    """
    A coordinate or frame in the Heliocentric system, which is observer-based.

    - The origin is the center of the Sun.
    - The Z-axis is aligned with the Sun-observer line.
    - The Y-axis is aligned with the component of the vector to the Sun's north pole that is
      perpendicular to the Z-axis.

    This frame defaults to a Cartesian component representation, which is known as Heliocentric
    Cartesian (HCC). This frame can also be represented using cylindrical components, where
    where ``rho`` is the impact parameter and ``psi`` is the position angle.
    ``psi`` is measured relative to the west limb, rather than solar north, so is shifted
    by 90 degrees compared to the convention of the Heliocentric Radial (HCR) system.

    A new instance can be created using the following signatures
    (note that if supplied, ``obstime``, ``observer``, and ``representation_type`` must be
    keyword arguments)::

        Heliocentric(x, y, z, obstime=obstime, observer=observer)
        Heliocentric(rho, psi, z, representation_type='cylindrical', obstime=obstime, observer=observer)

    Parameters
    ----------
    {data}
    {xyz}
    {observer}
    {common}

    Examples
    --------

    >>> from astropy.coordinates import SkyCoord, CartesianRepresentation
    >>> import sunpy.coordinates
    >>> import astropy.units as u

    >>> sc = SkyCoord(CartesianRepresentation(10*u.km, 1*u.km, 2*u.km),
    ...               obstime="2011/01/05T00:00:50", observer="earth", frame="heliocentric")
    >>> sc
    <SkyCoord (Heliocentric: obstime=2011-01-05T00:00:50.000, observer=<HeliographicStonyhurst Coordinate for 'earth'>): (x, y, z) in km
        (10., 1., 2.)>

    >>> sc = SkyCoord([1,2]*u.km, [3,4]*u.m, [5,6]*u.cm,
    ...               obstime="2011/01/01T00:00:54", observer="earth", frame="heliocentric")
    >>> sc
    <SkyCoord (Heliocentric: obstime=2011-01-01T00:00:54.000, observer=<HeliographicStonyhurst Coordinate for 'earth'>): (x, y, z) in (km, m, cm)
        [(1., 3., 5.), (2., 4., 6.)]>

    >>> sc = SkyCoord(CylindricalRepresentation(10*u.km, 60*u.deg, 10*u.km),
    ...               obstime="2011/01/05T00:00:50", observer="earth", frame="heliocentric")
    >>> sc
    <SkyCoord (Heliocentric: obstime=2011-01-05T00:00:50.000, observer=<HeliographicStonyhurst Coordinate for 'earth'>): (x, y, z) in km
        (5., 8.66025404, 10.)>
    """
    default_representation = CartesianRepresentation
    default_differential = CartesianDifferential

    frame_specific_representation_info = {
        CylindricalRepresentation: [RepresentationMapping('phi', 'psi', u.deg)]
    }

    observer = ObserverCoordinateAttribute(HeliographicStonyhurst)

    def represent_as(self, base, s='base', in_frame_units=False):
        data = super().represent_as(base, s, in_frame_units=in_frame_units)

        # For cylindrical representations, wrap the `psi` component (natively `phi`) at 360 deg
        if isinstance(data, CylindricalRepresentation):
            data.phi.wrap_at(360*u.deg, inplace=True)
        return data


@add_common_docstring(**_frame_parameters())
class Helioprojective(SunPyBaseCoordinateFrame):
    """
    A coordinate or frame in the Helioprojective Cartesian (HPC) system.

    This is an observer-based spherical coordinate system, with:

    - The origin is the observer location.
    - The line connecting the poles of the frame is parallel to the component of the Sun's
      rotation axis that is perpendicular to the observer-Sun line.
    - ``Tx`` (short for "theta_x", or :math:`\\theta_x`) is the longitude, the angle relative to
      the plane containing the observer-Sun line and the poles of the frame, with positive values
      in the direction of the Sun's west limb.
    - ``Ty`` (short for "theta_y", or :math:`\\theta_y`) is the latitude, the angle relative to the
      plane that is perpendicular to the poles of the frame, with positive values in the direction
      of the Sun's north pole.
    - ``distance`` is the observer-object distance.

    .. note::
        It can be confusing that the name of this coordinate system uses the adjective "Cartesian"
        despite being a spherical coordinate system. The reason for this name is because close to
        the center of the Sun where the small-angle approximation is appropriate, ``Tx`` and ``Ty``
        can be thought of as two components of a Cartesian-like coordinate system. The corresponding
        third Cartesian-like component, sometimes called zeta (:math:`\\zeta`), is defined to be the
        Sun-observer distance (``observer.radius``) minus the observer-object distance (``distance``).

    This system is frequently used in a projective form without ``distance`` specified. When
    transforming such a 2D coordinate to another frame, the object location usually needs to be
    fully 3D, which is achieved by calling :meth:`.make_3d` to generate the ``distance`` component.
    The default assumption is that the object lies on the surface of the Sun if the 2D coordinate is
    on the solar disk, but is otherwise undefined if the 2D coordinate is beyond the solar limb.
    This assumption can be modified using a screen (e.g., :func:`~sunpy.coordinates.SphericalScreen`).

    A new instance can be created using the following signatures
    (note that if supplied, ``obstime`` and ``observer`` must be keyword arguments)::

        Helioprojective(Tx, Ty, obstime=obstime, observer=observer)
        Helioprojective(Tx, Ty, distance, obstime=obstime, observer=observer)

    Parameters
    ----------
    {data}
    Tx : `~astropy.coordinates.Angle` or `~astropy.units.Quantity`
        The theta_x (:math:`\\theta_x`) component for this object. Not needed if ``data`` is given.
    Ty : `~astropy.coordinates.Angle` or `~astropy.units.Quantity`
        The theta_y (:math:`\\theta_y`) component for this object. Not needed if ``data`` is given.
    distance : `~astropy.units.Quantity`
        The distance component from the observer for this object. Not needed if ``data`` is given.
    {observer}
    {rsun}
    {common}

    See Also
    --------
    HelioprojectiveRadial
    ~sunpy.coordinates.PlanarScreen, ~sunpy.coordinates.SphericalScreen

    Examples
    --------
    >>> from astropy.coordinates import SkyCoord
    >>> import sunpy.coordinates
    >>> import astropy.units as u
    >>> sc = SkyCoord(0*u.deg, 0*u.deg, 5*u.km,
    ...               obstime="2010/01/01T00:00:00", observer="earth", frame="helioprojective")
    >>> sc
    <SkyCoord (Helioprojective: obstime=2010-01-01T00:00:00.000, rsun=695700.0 km, observer=<HeliographicStonyhurst Coordinate for 'earth'>): (Tx, Ty, distance) in (arcsec, arcsec, km)
        (0., 0., 5.)>
    >>> sc = SkyCoord(0*u.deg, 0*u.deg,
    ...               obstime="2010/01/01T00:00:00", observer="earth", frame="helioprojective")
    >>> sc
    <SkyCoord (Helioprojective: obstime=2010-01-01T00:00:00.000, rsun=695700.0 km, observer=<HeliographicStonyhurst Coordinate for 'earth'>): (Tx, Ty) in arcsec
        (0., 0.)>
    >>> sc = SkyCoord(CartesianRepresentation(1*u.AU, 1e5*u.km, -2e5*u.km),
    ...               obstime="2011/01/05T00:00:50", observer="earth", frame="helioprojective")
    >>> sc
    <SkyCoord (Helioprojective: obstime=2011-01-05T00:00:50.000, rsun=695700.0 km, observer=<HeliographicStonyhurst Coordinate for 'earth'>): (Tx, Ty, distance) in (arcsec, arcsec, AU)
        (137.87948623, -275.75878762, 1.00000112)>
    """
    frame_specific_representation_info = {
        SphericalRepresentation: [RepresentationMapping('lon', 'Tx', u.arcsec),
                                  RepresentationMapping('lat', 'Ty', u.arcsec),
                                  RepresentationMapping('distance', 'distance', None)],
        SphericalDifferential: [RepresentationMapping('d_lon', 'd_Tx', u.arcsec/u.s),
                                RepresentationMapping('d_lat', 'd_Ty', u.arcsec/u.s),
                                RepresentationMapping('d_distance', 'd_distance', u.km/u.s)],
        UnitSphericalRepresentation: [RepresentationMapping('lon', 'Tx', u.arcsec),
                                      RepresentationMapping('lat', 'Ty', u.arcsec)],
    }

    rsun = QuantityAttribute(default=_RSUN, unit=u.km)
    observer = ObserverCoordinateAttribute(HeliographicStonyhurst)
    _assumed_screen = None

    @property
    def angular_radius(self):
        """
        Angular radius of the Sun as seen by the observer.

        The ``rsun`` frame attribute is the radius of the Sun in length units.
        The tangent vector from the observer to the edge of the Sun forms a
        right-angle triangle with the radius of the Sun as the far side and the
        Sun-observer distance as the hypotenuse. Thus, the sine of the angular
        radius of the Sun is ratio of these two distances.
        """
        from sunpy.coordinates.sun import _angular_radius  # avoiding a circular import

        if not isinstance(self.observer, HeliographicStonyhurst):
            if self.observer is None:
                raise ValueError("The observer must be defined, not `None`.")
            raise ValueError("The observer must be fully defined by specifying `obstime`.")
        return _angular_radius(self.rsun, self.observer.radius)

    def make_3d(self):
        """
        This method calculates the third coordinate of the Helioprojective
        frame. It assumes that the coordinate point is on the surface of the Sun.

        If a point in the frame is off limb then NaN will be returned.

        Returns
        -------
        new_frame : `~sunpy.coordinates.frames.Helioprojective`
            A new frame instance with all the attributes of the original but
            now with a third coordinate.
        """
        # Skip if we already are 3D
        if not self._is_2d:
            return self

        if not isinstance(self.observer, BaseCoordinateFrame):
            raise ConvertError("Cannot calculate distance to the Sun "
                               f"for observer '{self.observer}' "
                               "without `obstime` being specified.")

        rep = self.represent_as(UnitSphericalRepresentation)
        lat, lon = rep.lat, rep.lon

        # Check for the use of floats with lower precision than the native Python float
        if not set([lon.dtype.type, lat.dtype.type]).issubset([float, np.float64, np.longdouble]):
            warn_user("The Helioprojective component values appear to be lower "
                      "precision than the native Python float: "
                      f"Tx is {lon.dtype.name}, and Ty is {lat.dtype.name}. "
                      "To minimize precision loss, you may want to cast the values to "
                      "`float` or `numpy.float64` via the NumPy method `.astype()`.")

        # Calculate the distance to the surface of the Sun using the law of cosines
        cos_alpha = np.cos(lat) * np.cos(lon)
        c = self.observer.radius**2 - self.rsun**2
        b = -2 * self.observer.radius * cos_alpha
        # Ignore sqrt of NaNs
        with np.errstate(invalid='ignore'):
            d = ((-1*b) - np.sqrt(b**2 - 4*c)) / 2  # use the "near" solution

        if self._assumed_screen:
            d_screen = self._assumed_screen.calculate_distance(self)
            d = np.fmin(d, d_screen) if self._assumed_screen.only_off_disk else d_screen

        # This warning can be triggered in specific draw calls when plt.show() is called
        # we can not easily prevent this, so we check the specific function is being called
        # within the stack trace.
        stack_trace = traceback.format_stack()
        matching_string = 'wcsaxes.*(_draw_grid|_update_ticks)'
        bypass = any([re.search(matching_string, string) for string in stack_trace])
        if not bypass and np.all(np.isnan(d)) and np.any(np.isfinite(cos_alpha)):
            warn_user("The conversion of these 2D helioprojective coordinates to 3D is all NaNs "
                      "because off-disk coordinates need an additional assumption to be mapped to "
                      "calculate distance from the observer. Consider using the context manager "
                      "`SphericalScreen()`.")

        return self.realize_frame(SphericalRepresentation(lon=lon,
                                                          lat=lat,
                                                          distance=d))

    @u.quantity_input
    def is_visible(self, *, tolerance: u.m = 1*u.m):
        """
        Returns whether the coordinate is on the visible side of the Sun.

        A coordinate is visible if it can been seen from the observer (the ``observer``
        frame attribute) assuming that the Sun is an opaque sphere with a fixed radius
        (the ``rsun`` frame attribute). The visible side of the Sun is always smaller
        than a full hemisphere.

        Parameters
        ----------
        tolerance : `~astropy.units.Quantity`
            The depth below the surface of the Sun that should be treated as
            transparent.

        Notes
        -----
        If the coordinate is 2D, it is automatically deemed visible. A 2D coordinate
        describes a look direction from the observer, who would simply see whatever is
        in "front", and thus cannot correspond to a point hidden from the observer.

        The ``tolerance`` parameter accommodates situations where the limitations of
        numerical precision would falsely conclude that a coordinate is not visible.
        For example, a coordinate that is expressly created to be on the solar surface
        may be calculated to be slightly below the surface, and hence not visible if
        there is no tolerance. However, a consequence of the ``tolerance`` parameter
        is that a coordinate that is formally on the far side of the Sun but is
        extremely close to the solar limb can be evaluated as visible. With the
        default ``tolerance`` value of 1 meter, a coordinate on the surface of the Sun
        can be up to 11 arcseconds of heliographic longitude past the solar limb and
        still be evaluated as visible.

        Examples
        --------
        >>> import numpy as np
        >>> import astropy.units as u
        >>> from astropy.coordinates import SkyCoord
        >>> from sunpy.coordinates import HeliographicStonyhurst, Helioprojective

        >>> hpc_frame = Helioprojective(observer='earth', obstime='2023-08-03')

        >>> in_front = SkyCoord(0*u.arcsec, 0*u.arcsec, 0.5*u.AU, frame=hpc_frame)
        >>> print(in_front.is_visible())
        True

        >>> behind = SkyCoord(0*u.arcsec, 0*u.arcsec, 1.5*u.AU, frame=hpc_frame)
        >>> print(behind.is_visible())
        False

        >>> hgs_array = SkyCoord(np.arange(-180, 180, 60)*u.deg, [0]*6*u.deg,
        ...                      frame='heliographic_stonyhurst', obstime='2023-08-03')
        >>> print(hgs_array)
        <SkyCoord (HeliographicStonyhurst: obstime=2023-08-03T00:00:00.000, rsun=695700.0 km): (lon, lat) in deg
            [(-180., 0.), (-120., 0.), ( -60., 0.), (   0., 0.), (  60., 0.),
             ( 120., 0.)]>
        >>> print(hgs_array.transform_to(hpc_frame).is_visible())
        [False False  True  True  True False]
        """
        # If the coordinate is 2D, it must be visible
        if self._is_2d:
            return np.ones_like(self.data, dtype=bool)

        # Use a slightly smaller solar radius to accommodate numerical precision
        solar_radius = self.rsun - tolerance

        data = self.cartesian
        data_to_sun = self.observer.radius * CartesianRepresentation(1, 0, 0) - data

        # When representing the helioprojective point as true Cartesian, the X value is the
        # distance from the observer to the point in the sunward direction
        is_behind_observer = data.x < 0
        # When comparing heliocentric angles, we compare the sine values and hence avoid calling arcsin()
        is_beyond_limb = np.sqrt(data.y **2 + data.z **2) / data.norm() > solar_radius / self.observer.radius
        is_above_surface = data_to_sun.norm() >= solar_radius
        is_on_near_side = data.dot(data_to_sun) >= 0

        return is_behind_observer | is_beyond_limb | (is_on_near_side & is_above_surface)


@add_common_docstring(**_frame_parameters())
class HelioprojectiveRadial(SunPyBaseCoordinateFrame):
    """
    A coordinate or frame in the Helioprojective Radial system.

    This is an observer-based spherical coordinate system, with:

    - ``psi`` is the position angle of the coordinate, measured eastward from solar
      north
    - ``delta`` is the declination angle, which is the impact angle (the angle
      between the observer-Sun line and the observer-coordinate line) minus 90
      degrees
    - ``r`` is the observer-coordinate distance

    .. note::
        The declination angle, rather than the impact angle, is used as a component
        in order to match the FITS WCS definition. The impact angle can be readily
        retrieved using the `theta` property.

    Parameters
    ----------
    {data}
    psi : `~astropy.coordinates.Angle` or `~astropy.units.Quantity`
        The position angle. Not needed if ``data`` is given.
    delta : `~astropy.coordinates.Angle` or `~astropy.units.Quantity`
        The declination angle. Not needed if ``data`` is given.
    r: `~astropy.coordinates.Angle` or `~astropy.units.Quantity`
        The observer-coordinate distance. Not needed if ``data`` is given.
    {rsun}
    {observer}
    {common}

    See Also
    --------
    Helioprojective

    Examples
    --------
    >>> from astropy.coordinates import SkyCoord
    >>> import sunpy.coordinates
    >>> import astropy.units as u

    >>> sc = SkyCoord(0*u.deg, -90*u.deg, 5*u.km,
    ...               obstime="2010/01/01T00:00:00", observer="earth", frame="helioprojectiveradial")
    >>> sc
    <SkyCoord (HelioprojectiveRadial: obstime=2010-01-01T00:00:00.000, rsun=695700.0 km, observer=<HeliographicStonyhurst Coordinate for 'earth'>): (psi, delta, distance) in (deg, deg, km)
        (0., -90., 5.)>
    >>> sc.theta
    <Angle 0. arcsec>

    >>> sc = SkyCoord(30*u.deg, -89.9*u.deg,
    ...               obstime="2010/01/01T00:00:00", observer="earth", frame="helioprojectiveradial")
    >>> sc
    <SkyCoord (HelioprojectiveRadial: obstime=2010-01-01T00:00:00.000, rsun=695700.0 km, observer=<HeliographicStonyhurst Coordinate for 'earth'>): (psi, delta) in deg
        (30., -89.9)>
    >>> sc.theta
    <Angle 360. arcsec>

    >>> sc = SkyCoord(CartesianRepresentation(1e5*u.km, -2e5*u.km, -1*u.AU),
    ...               obstime="2011/01/05T00:00:50", observer="earth", frame="helioprojectiveradial")
    >>> sc
    <SkyCoord (HelioprojectiveRadial: obstime=2011-01-05T00:00:50.000, rsun=695700.0 km, observer=<HeliographicStonyhurst Coordinate for 'earth'>): (psi, delta, distance) in (deg, deg, km)
        (296.56505118, -89.91435897, 1.49598038e+08)>
    >>> sc.theta
    <Angle 308.30772022 arcsec>

    .. minigallery:: sunpy.coordinates.HelioprojectiveRadial
    """
    _wrap_angle = 360*u.deg

    default_representation = SphericalRepresentation

    frame_specific_representation_info = {
        SphericalRepresentation: [RepresentationMapping('lon', 'psi', u.deg),
                                  RepresentationMapping('lat', 'delta', u.deg),
                                  RepresentationMapping('distance', 'distance', None)],
        SphericalDifferential: [RepresentationMapping('d_lon', 'd_psi', u.deg/u.s),
                                RepresentationMapping('d_lat', 'd_delta', u.deg/u.s),
                                RepresentationMapping('d_distance', 'd_distance', u.km/u.s)],
        UnitSphericalRepresentation: [RepresentationMapping('lon', 'psi', u.deg),
                                      RepresentationMapping('lat', 'delta', u.deg)],
    }

    rsun = QuantityAttribute(default=_RSUN, unit=u.km)
    observer = ObserverCoordinateAttribute(HeliographicStonyhurst)

    @property
    def theta(self):
        """
        Returns the impact angle, which is the declination angle plus 90 degrees.
        """
        return (90*u.deg + self.spherical.lat).to(u.arcsec)

    def make_3d(self):
        """
        Returns a 3D version of this coordinate.

        If the coordinate is 2D, the default assumption is that the coordinate is on
        the surface of the Sun, and the distance component is calculated
        accordingly. Under this assumption, if the 2D coordinate is outside the
        disk, the distance component will be NaN.

        The assumption can be changed using one of the screens in
        `sunpy.coordinates.screens`.

        Returns
        -------
        `~sunpy.coordinates.frames.HelioprojectiveRadial`
            The 3D version of this coordinate.
        """
        # Skip if we already are 3D
        if not self._is_2d:
            return self

        # Make 3D by going through HPC, which thus will make use of any screen
        hpc_frame = Helioprojective(obstime=self.obstime, observer=self.observer, rsun=self.rsun)
        return self.transform_to(hpc_frame).make_3d().transform_to(self)


@add_common_docstring(**_frame_parameters())
class HeliocentricEarthEcliptic(SunPyBaseCoordinateFrame):
    """
    A coordinate or frame in the Heliocentric Earth Ecliptic (HEE) system.

    - The origin is the center of the Sun.
    - The X-axis (0 degrees longitude and 0 degrees latitude) is aligned with the Sun-Earth line.
    - The Z-axis (+90 degrees latitude) is aligned with the component perpendicular to the X-axis
      of the mean ecliptic pole at the observation time.

    Parameters
    ----------
    {data}
    {lonlat}
    {distance_sun}
    {common}
    """


@add_common_docstring(**_frame_parameters())
class GeocentricSolarEcliptic(SunPyBaseCoordinateFrame):
    """
    A coordinate or frame in the Geocentric Solar Ecliptic (GSE) system.

    - The origin is the center of the Earth.
    - The X-axis (0 degrees longitude and 0 degrees latitude) is aligned with the Earth-Sun line.
    - The Z-axis (+90 degrees latitude) is aligned with the component perpendicular to the X-axis
      of the mean ecliptic pole at the observation time.

    Parameters
    ----------
    {data}
    {lonlat}
    {distance_earth}
    {common}

    Notes
    -----
    Aberration due to Earth motion is not included when transforming directly to/from this frame.
    """


@add_common_docstring(**_frame_parameters())
class HeliocentricInertial(SunPyBaseCoordinateFrame):
    """
    A coordinate or frame in the Heliocentric Inertial (HCI) system.

    - The origin is the center of the Sun.
    - The Z-axis (+90 degrees latitude) is aligned with the Sun's north pole.
    - The X-axis (0 degrees longitude and 0 degrees latitude) is aligned with the solar ascending
      node on the ecliptic (mean J2000.0).

    Parameters
    ----------
    {data}
    {lonlat}
    {distance_sun}
    {common}

    Notes
    -----
    The solar ascending node on the ecliptic lies on the intersection of the solar equatorial
    plane with the ecliptic plane, not on the intersection of the celestial equatorial plane with
    the ecliptic plane.
    """


@add_common_docstring(**_frame_parameters())
class GeocentricEarthEquatorial(SunPyBaseCoordinateFrame):
    """
    A coordinate or frame in the Geocentric Earth Equatorial (GEI) system.

    - The origin is the center of the Earth.
    - The Z-axis (+90 degrees latitude) is aligned with the Earth's mean (not true) north pole.
    - The X-axis (0 degrees longitude and 0 degrees latitude) is aligned with the mean (not true)
      vernal equinox.

    Parameters
    ----------
    {data}
    {lonlat}
    {distance_earth}
    {equinox}
    {common}

    Notes
    -----
    Aberration due to Earth motion is not included when transforming directly to/from this frame.
    """
    equinox = TimeFrameAttributeSunPy(default=_J2000)


class BaseMagnetic(SunPyBaseCoordinateFrame):
    """
    Base class for frames that rely on the Earth's magnetic model (MAG, SM, and GSM).

    This class is not intended to be used directly and has no transformations defined.
    """
    magnetic_model = Attribute(default='igrf13')

    @property
    def _igrf_file(self):
        if not self.magnetic_model.startswith("igrf"):
            raise ValueError

        # First look if the file is bundled in package
        local_file = os.path.join(os.path.dirname(__file__), "data",
                                  f"{self.magnetic_model}coeffs.txt")
        if os.path.exists(local_file):
            return local_file

        # Otherwise download the file and cache it
        return download_file("https://www.ngdc.noaa.gov/IAGA/vmod/coeffs/"
                             f"{self.magnetic_model}coeffs.txt", cache=True)

    @property
    def _lowest_igrf_coeffs(self):
        with open(self._igrf_file) as f:
            while not (line := f.readline()).startswith('g/h'):
                pass

            years = list(map(float, line.split()[3:-1]))
            g10s = list(map(float, f.readline().split()[3:]))
            g11s = list(map(float, f.readline().split()[3:]))
            h11s = list(map(float, f.readline().split()[3:]))

        decimalyear = self.obstime.utc.decimalyear
        output_shape = decimalyear.shape
        if np.any(decimalyear < 1900.0):
            raise ValueError("At least one of the dates is earlier than the year 1900, which is unsupported.")

        # Use piecewise linear interpolation before the last year
        decimalyear = np.atleast_1d(decimalyear)
        g10 = np.interp(decimalyear, years, g10s[:-1])
        g11 = np.interp(decimalyear, years, g11s[:-1])
        h11 = np.interp(decimalyear, years, h11s[:-1])

        # Use secular variation beyond the last year
        future = decimalyear > years[-1]
        if np.sum(future) > 0:
            g10[future] = g10s[-2] + (decimalyear[future] - years[-1]) * g10s[-1]
            g11[future] = g11s[-2] + (decimalyear[future] - years[-1]) * g11s[-1]
            h11[future] = h11s[-2] + (decimalyear[future] - years[-1]) * h11s[-1]

        g10 = g10.reshape(output_shape)
        g11 = g11.reshape(output_shape)
        h11 = h11.reshape(output_shape)

        return g10, g11, h11

    @add_common_docstring(**_frame_parameters())
    @property
    def dipole_lonlat(self):
        """
        The geographic longitude/latitude of the Earth's magnetic north pole.

        This position is calculated from the first three coefficients of the selected
        IGRF model per Franz & Harper (2002). The small offset between dipole center
        and Earth center is ignored.

        References
        ----------
        {igrf_reference}
        """
        g10, g11, h11 = self._lowest_igrf_coeffs
        # Intentionally use arctan() instead of arctan2() to get angles in specific quadrants
        lon = (np.arctan(h11 / g11) << u.rad).to(u.deg)
        lat = 90*u.deg - np.arctan((g11 * np.cos(lon) + h11 * np.sin(lon)) / g10)
        return Longitude(lon, wrap_angle=180*u.deg), Latitude(lat)

    @add_common_docstring(**_frame_parameters())
    @property
    def dipole_moment(self):
        """
        The Earth's dipole moment.

        The moment is calculated from the first three coefficients of the selected
        IGRF model per Franz & Harper (2002).

        References
        ----------
        {igrf_reference}
        """
        g10, g11, h11 = self._lowest_igrf_coeffs
        moment = np.sqrt(g10**2 + g11**2 + h11**2) * R_earth**3
        return moment


@add_common_docstring(**_frame_parameters())
class Geomagnetic(BaseMagnetic):
    """
    A coordinate or frame in the Geomagnetic (MAG) system.

    - The origin is the center of the Earth.
    - The Z-axis (+90 degrees latitude) is aligned with the Earth's magnetic north
      pole.
    - The X-axis (0 degrees longitude and 0 degrees latitude) is aligned with the
      component of the Earth's geographic north pole that is perpendicular to the
      Z-axis.

    Parameters
    ----------
    {data}
    {lonlat}
    {distance_earth}
    {magnetic_model}
    {common}

    Notes
    -----
    The position of Earth's magnetic north pole is calculated from the first three
    coefficients of the selected IGRF model per Franz & Harper (2002). The small
    offset between dipole center and Earth center is ignored.

    References
    ----------
    {igrf_reference}
    """


@add_common_docstring(**_frame_parameters())
class SolarMagnetic(BaseMagnetic):
    """
    A coordinate or frame in the Solar Magnetic (SM) system.

    - The origin is the center of the Earth.
    - The Z-axis (+90 degrees latitude) is aligned with the Earth's magnetic north
      pole.
    - The X-axis (0 degrees longitude and 0 degrees latitude) is aligned with the
      component of the Earth-Sun line that is perpendicular to the Z-axis.

    Parameters
    ----------
    {data}
    {lonlat}
    {distance_earth}
    {magnetic_model}
    {common}

    Notes
    -----
    The position of Earth's magnetic north pole is calculated from the first three
    coefficients of the selected IGRF model per Franz & Harper (2002). The small
    offset between dipole center and Earth center is ignored.

    References
    ----------
    {igrf_reference}
    """


@add_common_docstring(**_frame_parameters())
class GeocentricSolarMagnetospheric(BaseMagnetic):
    """
    A coordinate or frame in the GeocentricSolarMagnetospheric (GSM) system.

    - The origin is the center of the Earth.
    - The X-axis (0 degrees longitude and 0 degrees latitude) is aligned with the
      Earth-Sun line.
    - The Z-axis (+90 degrees latitude) is aligned with the component of the Earth's
      magnetic north pole that is perpendicular to the X-axis.

    Parameters
    ----------
    {data}
    {lonlat}
    {distance_earth}
    {magnetic_model}
    {common}

    Notes
    -----
    The position of Earth's magnetic north pole is calculated from the first three
    coefficients of the selected IGRF model per Franz & Harper (2002). The small
    offset between dipole center and Earth center is ignored.

    References
    ----------
    {igrf_reference}
    """



"""
Provides classes for creating mocked objects.
"""
import io
from collections import defaultdict
from collections.abc import MutableMapping

__all__ = ["MockObject", "MockHTTPResponse", "MockOpenTextFile"]


class MockObject(MutableMapping):
    """
    Object from which we can construct other "mocked" objects.

    Limitations
    -----------
    On initiation a `ValueError` will be raised if any of the ``kwargs`` have the same name as an
    existing attribute/method of the `~sunpy.tests.mock.MockObject` or underlying data store.

    Updating existing attributes, or adding new ones, should only be done using bracket
    notation and **not** dot notation.
    Using dot notation will update the `~sunpy.tests.mock.MockObject` and not the data store.

    Examples
    --------
    >>> mo = MockObject(code=400)
    >>> mo.code
    400
    >>> mo['code']
    400
    >>> mo.code = 210
    >>> mo.code
    210
    >>> mo['code']
    400

    The recommended way of changing the value of an existing, or new, attribute is using bracket notation:

    >>> m = MockObject(start='now')
    >>> m['start'] = 'Thursday'
    >>> m['start']
    'Thursday'
    >>> m.start
    'Thursday'
    """

    def __init__(self, *args, **kwargs):
        self._datastore = dict()
        self.prohibited_attrs = set(dir(self))
        self.prohibited_attrs.update(dir(self._datastore))

        for candidate in kwargs.keys():
            if candidate in self.prohibited_attrs:
                raise ValueError(f"kwarg '{candidate}' is already an attribute "
                                 f"of {type(self._datastore)} or {type(self)}")
        self.update(dict(*args, **kwargs))

    def __getattr__(self, name):
        if name in self._datastore:
            return self._datastore[name]

        raise AttributeError(name)

    def __getitem__(self, name):
        return self._datastore[name]

    def __setitem__(self, name, value):
        if name in self.prohibited_attrs:
            raise ValueError(f"Name '{name}' is already an attribute "
                             f"of {type(self._datastore)} or {type(self)}")
        self._datastore[name] = value

    def __delitem__(self, name):
        raise NotImplementedError(f"'del' operation for {self.__class__.__name__} "
                                  "not supported")

    def __iter__(self):
        return iter(self._datastore)

    def __len__(self):
        return len(self._datastore)

    def __repr__(self):
        return (f"<{self.__module__}.{self.__class__.__name__} {self._datastore} at {hex(id(self))}>")


class MockHTTPResponse(MockObject):
    """
    The result of calling `~urllib.request.urlopen`. For this implementation we
    are only interested in querying the "headers" attribute, which is a
    http.client.HTTPMessage object.

    Parameters
    ----------
    url : `str` optional
        The url of the connection. Defaults to ''.
    headers : `dict` of `str` optional
        HTTP header fields of the response message. Defaults to ``{}``.

    Limitations
    -----------
    On a "real" ``http.client.HTTPMessage``, header name retrieval is case insensitive.
    In this implementation the header names are case sensitive.

    Examples
    --------
    >>> result = MockHTTPResponse(url='https://abc.com', headers={'Content-Type':'text/html'})
    >>> result.headers.get('Content-Type')
    'text/html'
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setdefault('url', '')

        headers_store = defaultdict(lambda: None)

        if 'headers' in self:
            headers_store.update(self['headers'])

        self['headers'] = headers_store


class MockOpenTextFile(MockObject):
    """
    Partial implementation of a file like object for reading/writing text
    files.

    Binary files are **not** supported.

    Parameters
    ----------
    file : `str`, optional
        The name of the file. Default to "N/A".
    mode : `str`, optional, default:'r'
        The way in which the file is to be used. Defaults to "r", i.e., read.
    data : `str`, optional
        The initial data which can be read from the file. Defaults to ''.

    Limitations
    -----------
    Unlike in a real file, this implementation makes no attempt to keep
    track of where we are, when reading or writing.

    Examples
    --------
    >>> dummy_read_only = MockOpenTextFile()
    >>> named_write = MockOpenTextFile(file='a.txt', mode='w')
    >>> named_read = MockOpenTextFile('b.txt')
    >>> named_rd_wr = MockOpenTextFile('c.txt', 'r+', data='Hello, world')
    """

    def __init__(self, *args, **kwargs):
        # Positional and/or keyword args can be used for the 'file' & 'mode'
        # parameters. Could do a lot more checking to make sure all required
        # arguments are present

        num_pos_args = len(args)

        if num_pos_args == 1:
            kwargs['file'] = args[0]
        elif num_pos_args == 2:
            kwargs['file'] = args[0]
            kwargs['mode'] = args[1]

        if 'mode' not in kwargs:
            kwargs['mode'] = 'r'

        super().__init__(**kwargs)
        self.setdefault('file', 'N/A')
        self['name'] = self['file']
        self.setdefault('closed', False)
        self.setdefault('data', '')

    def write(self, content):
        if not self.writable():
            raise io.UnsupportedOperation(':not writable')

        self.data += content

        return len(content)

    def read(self):
        if not self.readable():
            raise io.UnsupportedOperation(': not readable')

        return self.data

    def readlines(self):
        if self.closed:
            raise ValueError('I/O operation on closed file')

        # Documentation recommends using '\n' as the line terminator when reading/writing text
        # files. See `os.linesep` in https://docs.python.org/3/library/os.html
        new_line = '\n'
        return [f'{line}{new_line}' for line in self.data.split(new_line)]

    def readable(self):
        if self.closed:
            raise ValueError('I/O operation on closed file')

        return 'r' in self.mode

    def writable(self):
        if self.closed:
            raise ValueError('I/O operation on closed file')

        return ('w' in self.mode) or ('r+' in self.mode)

    def close(self):
        self.closed = True
        self.data = ''

    def __repr__(self):
        return (f"<{self.__module__}.{self.__class__.__name__} file '{self.file}' mode '{self.mode}' "
                f"at {hex(id(self))}>")



"""
=============================
Plot positions on a blank map
=============================

This example showcases how to plot positions on a blank map.
It is often useful to plot coordinate positions of events on a blank helioprojective coordinate map.
In this example, we create an empty map with a WCS defined by a helioprojective frame as observed from
Earth at a certain time, and show how you can plot different coordinates on it.
"""
import matplotlib.pyplot as plt
import numpy as np

import astropy.units as u
from astropy.coordinates import SkyCoord

import sunpy.map
from sunpy.coordinates import frames

################################################################################
# First we will create a blank map using with an array of zeros.
# Since there is no WCS information, we will need to construct a header to pass to Map.

data = np.full((10, 10), np.nan)

# Define a reference coordinate and create a header using sunpy.map.make_fitswcs_header
skycoord = SkyCoord(0*u.arcsec, 0*u.arcsec, obstime='2013-10-28',
                    observer='earth', frame=frames.Helioprojective)

# Scale set to the following for solar limb to be in the field of view
header = sunpy.map.make_fitswcs_header(data, skycoord, scale=[220, 220]*u.arcsec/u.pixel)

# Use sunpy.map.Map to create the blank map
blank_map = sunpy.map.Map(data, header)

################################################################################
# Now we have constructed the map, we can plot it and mark important locations to it.
# Initialize the plot and add the map to it

# sphinx_gallery_defer_figures

fig = plt.figure()
ax = fig.add_subplot(projection=blank_map)
blank_map.plot(axes=ax)
blank_map.draw_limb(axes=ax, color="k")
blank_map.draw_grid(axes=ax, color="k")

################################################################################
# Coordinates that are being plotted - (0, 0), (50, 100) and (400, 400).

# sphinx_gallery_defer_figures

xc = [0, 50, 400] * u.arcsec
yc = [0, 100, 400] * u.arcsec

################################################################################
# Plot the blank map with the specified coordinates. Note that the marker for
# (0, 0) in helioprojective coordinates is at the center of the solar disk, yet
# the heliographic equator (zero degrees latitude) does not go through the disk
# center and instead curves below it. The reason for that is the observer,
# specified as Earth in this example, is almost always at non-zero heliographic
# latitude, and disk center as seen by such an observer will have that same
# heliographic latitude. The :func:`~sunpy.coordinates.sun.B0` function returns
# Earth's heliographic latitude at a specified time.

coords = SkyCoord(xc, yc, frame=blank_map.coordinate_frame)
ax.plot_coord(coords, 'o')
ax.set_title('Plotting fixed points on a blank map')

plt.show()



import pytest

import astropy.units as u

from sunpy.data.test import get_dummy_map_from_header, get_test_filepath
from sunpy.map.mapbase import SpatialPair
from sunpy.map.sources.trace import TRACEMap
from sunpy.util.exceptions import SunpyMetadataWarning
from .helpers import _test_private_date_setters


@pytest.fixture(scope="module")
def trace_map():
    return get_dummy_map_from_header(get_test_filepath("tsi20010130_025823_a2.header"))


def test_fitstoTRACE(trace_map):
    """Tests the creation of TRACEMap using FITS."""
    assert isinstance(trace_map, TRACEMap)


def test_trace_coordinate_system(trace_map):
    assert trace_map.coordinate_system ==  SpatialPair(axis1='HPLN-TAN', axis2='HPLT-TAN')


def test_is_datasource_for(trace_map):
    """Test the is_datasource_for method of TRACEMap.
    Note that header data to be provided as an argument
    can be a MetaDict object."""
    assert trace_map.is_datasource_for(trace_map.data, trace_map.meta)


def test_reference_date(trace_map):
    assert trace_map.reference_date.isot == "2001-01-30T02:58:23.429"


def test_date(trace_map):
    assert trace_map.date.isot == "2001-01-30T02:58:23.429"


def test_private_date_setters(trace_map):
    _test_private_date_setters(trace_map)


def test_measurement(trace_map):
    """Tests the measurement property of the TRACEMap object."""
    assert int(trace_map.measurement) == 171


def test_observatory(trace_map):
    """Tests the observatory property of the TRACEMap object."""
    assert trace_map.observatory == "TRACE"


def test_norm_clip(trace_map):
    # Tests that the default normalizer has clipping disabled
    assert not trace_map.plot_settings['norm'].clip


def test_wcs(trace_map):
    # Smoke test that WCS is valid and can transform from pixels to world coordinates
    with pytest.warns(SunpyMetadataWarning, match='Missing metadata for observer'):
        trace_map.pixel_to_world(0*u.pix, 0*u.pix)



"""
==================================
Visualizing 3D stereoscopic images
==================================

How to make an anaglyph 3D image from a stereoscopic observation

We use a stereoscopic observation from July 2023, when the STEREO-A spacecraft
and the SDO spacecraft were close in heliographic longitude. See the
`Wikipedia page on anaglyph 3D <https://en.wikipedia.org/wiki/Anaglyph_3D>`__,
which typically requires red-cyan glasses to visualize.
"""
import copy

import matplotlib.pyplot as plt
from matplotlib.colors import Normalize

import astropy.units as u
from astropy.coordinates import SkyCoord

import sunpy.data.sample
import sunpy.map
from sunpy.coordinates import SphericalScreen

##############################################################################
# Download co-temporal SDO/AIA image STEREO/EUVI images. The EUVI map does
# not explicitly define their reference radius of the Sun, so we set it to be
# the same as for the AIA map to silence some informational messages.

aia_map = sunpy.map.Map(sunpy.data.sample.AIA_STEREOSCOPIC_IMAGE)
euvi_map = sunpy.map.Map(sunpy.data.sample.EUVI_STEREOSCOPIC_IMAGE)
euvi_map.meta['rsun_ref'] = aia_map.meta['rsun_ref']


##############################################################################
# Verify that the angular separation between the two vantage points is a few
# degrees. If the angular separation much larger, the 3D effect will be bad.

print(euvi_map.observer_coordinate.separation(aia_map.observer_coordinate))

##############################################################################
# Define a convenience function to reproject an input map to an observer in
# same direction, but at a distance of 1 AU, and then reproject both maps. The
# purpose is to make the Sun exactly the same size in pixels in both maps.


def reproject_to_1au(in_map):
    header = sunpy.map.make_fitswcs_header(
        (1000, 1000),
        SkyCoord(
            0*u.arcsec, 0*u.arcsec,
            frame='helioprojective',
            obstime=in_map.date,
            observer=in_map.observer_coordinate.realize_frame(
                in_map.observer_coordinate.represent_as('unitspherical') * u.AU
            )
        ),
        scale=(2.2, 2.2)*u.arcsec/u.pixel
    )
    with SphericalScreen(in_map.observer_coordinate):
        return in_map.reproject_to(header)


euvi_map = reproject_to_1au(euvi_map)
aia_map = reproject_to_1au(aia_map)

##############################################################################
# Define linear scaling for the two images so that they look the same. The
# values here are empirical.

euvi_norm = Normalize(vmin=750, vmax=1e4, clip=True)
aia_norm = Normalize(vmin=0, vmax=2.1e3, clip=True)

##############################################################################
# Plot the two maps side by side. Those who are able to see 3D images by
# defocusing their eyes can see the 3D effect without the later anaglyph
# image.

fig = plt.figure(figsize=(8, 4))
fig.subplots_adjust(wspace=0)

ax1 = fig.add_subplot(121, projection=euvi_map)
euvi_map.plot(axes=ax1, cmap='gray', norm=euvi_norm)
ax1.grid(False)
ax1.set_title('STEREO/EUVI')

ax2 = fig.add_subplot(122, projection=aia_map)
aia_map.plot(axes=ax2, cmap='gray', norm=aia_norm)
ax2.coords[1].set_ticks_visible(False)
ax2.coords[1].set_ticklabel_visible(False)
ax2.grid(False)
ax2.set_title('SDO/AIA')

##############################################################################
# We will make a color anaglyph 3D image by creating two colormaps based on an
# initial colormap (here, ``'sdoaia171'``). The left-eye colormap is just the
# red channel, and the right-eye colormap is just the green and blue channels.
# Thus, we zero out the appropriate channels in the two colormaps.

cmap_left = copy.deepcopy(plt.get_cmap('sdoaia171'))
cmap_right = copy.deepcopy(cmap_left)

cmap_left._segmentdata['blue'] = [(0, 0, 0), (1, 0, 0)]
cmap_left._segmentdata['green'] = [(0, 0, 0), (1, 0, 0)]
cmap_right._segmentdata['red'] = [(0, 0, 0), (1, 0, 0)]

##############################################################################
# Finally, we build the anaglyph 3D image. To easily plot in RGB color
# channels, we will be passing a NxMx3 array to matplotlib's ``imshow``. That
# array needs to be manually normalized (to be between 0 and 1) and colorized
# (mapped into the 3 color channels). The colorizing actually adds a fourth
# layer for the alpha layer, so we discard that layer.

fig = plt.figure()
ax = fig.add_subplot(projection=aia_map)

euvi_data = euvi_norm(euvi_map.data)
aia_data = aia_norm(aia_map.data)
ax.imshow((cmap_left(euvi_data) + cmap_right(aia_data))[:, :, 0:3], origin='lower')

ax.set_xlabel('Helioprojective Longitude (Solar-X)')
ax.set_ylabel('Helioprojective Latitude (Solar-Y)')

plt.show()



import os
from pathlib import Path
from collections import defaultdict

import numpy
from extension_helpers import get_compiler
from setuptools import Extension


def get_extensions():
    if os.environ.get("SUNPY_NO_BUILD_ANA_EXTENSION"):
        return []
    cfg = defaultdict(list)
    cfg["include_dirs"].append(numpy.get_include())
    cfg["sources"].extend(list((Path(__file__).parent / "src" / "ana").glob("*.c")))
    if get_compiler() == 'msvc':
        cfg["extra_compile_args"].extend([
            "/O2",
            "/W3",
            "/utf-8",
        ])
        # quiet MSVC's CRT nags
        cfg["define_macros"].extend([
            ("_CRT_SECURE_NO_WARNINGS", None),
        ])
    else:
        cfg["extra_compile_args"].extend([
            "-std=c99",
            "-O3",
            "-w",
        ])

    return [Extension("sunpy.io._pyana", **cfg)]



import sunpy.timeseries
from sunpy.data.test import get_test_filepath
from sunpy.tests.helpers import figure_test

rhessi_filepath = get_test_filepath('hsi_obssumm_20120601_018_truncated.fits.gz')


def test_implicit_rhessi():
    # Test a RHESSI TimeSeries
    ts_rhessi = sunpy.timeseries.TimeSeries(rhessi_filepath)
    assert isinstance(ts_rhessi, sunpy.timeseries.sources.rhessi.RHESSISummaryTimeSeries)


def test_rhessi():
    # Test a RHESSI TimeSeries
    ts_rhessi = sunpy.timeseries.TimeSeries(rhessi_filepath, source='RHESSI')
    assert isinstance(ts_rhessi, sunpy.timeseries.sources.rhessi.RHESSISummaryTimeSeries)


def test_rhessi_plot_column(rhessi_test_ts):
    ax = rhessi_test_ts.plot(columns=['3 - 6 keV', '25 - 50 keV', '100 - 300 keV'])
    assert len(ax.lines) == 3
    assert '3 - 6 keV' == ax.lines[0].get_label()
    assert '25 - 50 keV' == ax.lines[1].get_label()
    assert '100 - 300 keV' == ax.lines[2].get_label()


@figure_test
def test_rhessi_peek(rhessi_test_ts):
    rhessi_test_ts.peek()



"""
==============
Cropping a Map
==============

How to crop a map by using submap.
"""
import matplotlib.pyplot as plt

import astropy.units as u
from astropy.coordinates import SkyCoord

import sunpy.data.sample
import sunpy.map

###############################################################################
# We start with the sample data

swap_map = sunpy.map.Map(sunpy.data.sample.SWAP_LEVEL1_IMAGE)

##############################################################################
# To crop the data you create a submap, specifying the top right and bottom
# left as SkyCoord objects.

top_right = SkyCoord(0 * u.arcsec, -200 * u.arcsec, frame=swap_map.coordinate_frame)
bottom_left = SkyCoord(-900 * u.arcsec, -900 * u.arcsec, frame=swap_map.coordinate_frame)
swap_submap = swap_map.submap(bottom_left, top_right=top_right)

###############################################################################
# Let's plot the results.

fig = plt.figure()
ax = fig.add_subplot(projection=swap_submap)
image = swap_submap.plot(axes=ax)
swap_submap.draw_limb(axes=ax)
swap_submap.draw_grid(axes=ax)

# Make some room and put the title at the top of the figure
ax.set_position([0.1, 0.1, 0.8, 0.7])
ax.set_title(ax.get_title(), pad=45)

plt.show()



import warnings

import numpy as np
import pytest
from matplotlib.figure import Figure

import astropy.units as u
from astropy.coordinates import SkyCoord
from astropy.wcs import WCS
from astropy.wcs.wcsapi import HighLevelWCSWrapper
from astropy.wcs.wcsapi.wrappers import SlicedLowLevelWCS

import sunpy.io
from sunpy.data.test import get_test_filepath
from sunpy.map import Map
from sunpy.tests.helpers import figure_test
from sunpy.visualization import drawing


@pytest.fixture
def aia171_test_map():
    return Map(get_test_filepath('aia_171_level1.fits'))


@pytest.fixture
def heliographic_test_map():
    (data, header), = sunpy.io._file_tools.read_file(get_test_filepath('heliographic_phase_map.fits.gz'))
    # Fix unit strings to prevent some astropy fits fixing warnings
    header['CUNIT1'] = 'deg'
    header['CUNIT2'] = 'deg'
    # Set observer location to avoid warnings later
    header['HGLN_OBS'] = 0.0
    return Map((data, header))


@figure_test
def test_draw_equator_aia171(aia171_test_map):
    fig = Figure()
    axes = fig.add_subplot(projection=aia171_test_map)
    aia171_test_map.plot(axes=axes)
    drawing.equator(axes)
    return fig


@figure_test
def test_draw_prime_meridian_aia171(aia171_test_map):
    fig = Figure()
    axes = fig.add_subplot(projection=aia171_test_map)
    aia171_test_map.plot(axes=axes)
    drawing.prime_meridian(axes)
    return fig


@figure_test
def test_heliographic_equator_prime_meridian(heliographic_test_map):
    fig = Figure()
    axes = fig.add_subplot(projection=heliographic_test_map)
    heliographic_test_map.plot(axes=axes)
    drawing.equator(axes, color="blue")
    drawing.prime_meridian(axes, color="red")
    return fig


def test_prime_meridian_error():
    axes = Figure().add_subplot(projection=WCS())
    with pytest.raises(ValueError, match="does not have an observer"):
        drawing.prime_meridian(axes)


def test_limb_invisible(aia171_test_map):
    aia_obs = aia171_test_map.observer_coordinate
    # Create a new observer on the opposite side of the Sun
    new_obs = SkyCoord(lon=aia_obs.lon + 180*u.deg,
                       lat=-aia_obs.lat,
                       radius=aia_obs.radius / 10,
                       frame=aia_obs.replicate_without_data())

    ax = Figure().add_subplot(projection=aia171_test_map)
    # Original observer
    visible, hidden = drawing.limb(ax, aia_obs)
    assert visible is not None
    assert hidden is None
    # Far side observer
    visible, hidden = drawing.limb(ax, new_obs)
    assert visible is None
    assert hidden is not None


@pytest.fixture
def cutout_wcs(aia171_test_map):
    header = {
        'WCSAXES': 2,
        'NAXIS1': 2000,
        'NAXIS2': 1000,
        'CRPIX1': 1000.5,
        'CRPIX2': 500.5,
        'CDELT1': 1.66666666666667E-04,
        'CDELT2': 1.66666666666667E-04,
        'CUNIT1': 'deg',
        'CUNIT2': 'deg',
        'CTYPE1': 'HPLN-TAN',
        'CTYPE2': 'HPLT-TAN',
        'CRVAL1': 0.0 ,
        'CRVAL2': 0.0 ,
        'LONPOLE': 180.0 ,
        'LATPOLE': 0.0 ,
        'DATE-OBS': '2011-02-15T00:00:01.340',
        'DSUN_OBS': aia171_test_map.wcs.wcs.aux.dsun_obs,
        # Ensures that at least part of the extent is behind the limb
        'HGLN_OBS': aia171_test_map.wcs.wcs.aux.hgln_obs-90.0,
        'HGLT_OBS': aia171_test_map.wcs.wcs.aux.hglt_obs-15.0
    }
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        return WCS(header=header)


@figure_test
def test_draw_extent(aia171_test_map, cutout_wcs):
    fig = Figure()
    ax = fig.add_subplot(projection=aia171_test_map)
    aia171_test_map.plot(axes=ax)
    drawing.extent(ax, cutout_wcs)
    return fig


@pytest.fixture
def wcs_3d_ln_lt_l_coupled(aia171_test_map):
    # WCS for a 3D data cube with two celestial axes and one wavelength axis.
    # The latitudinal dimension is coupled to the third pixel dimension through
    # a single off diagonal element in the PCij matrix
    header = {
        'NAXIS1': 100,
        'NAXIS2': 120,
        'NAXIS3': 50,
        'CTYPE1': 'HPLN-TAN',
        'CRPIX1': 5,
        'CDELT1': 5,
        'CUNIT1': 'arcsec',
        'CRVAL1': 0.0,

        'CTYPE2': 'HPLT-TAN',
        'CRPIX2': 5,
        'CDELT2': 5,
        'CUNIT2': 'arcsec',
        'CRVAL2': 0.0,

        'CTYPE3': 'UTC',
        'CRPIX3': 1.0,
        'CDELT3': 1,
        'CUNIT3': 's',
        'CRVAL3': 1.0,

        'PC1_1': 1,
        'PC1_2': 0,
        'PC1_3': 0,
        'PC2_1': 0,
        'PC2_2': 1,
        'PC2_3': -1.0,
        'PC3_1': 1.0,
        'PC3_2': 0.0,
        'PC3_3': 0.0,

        'WCSAXES': 3,

        'DATE-OBS': aia171_test_map.date.isot,
        'HGLN_OBS': aia171_test_map.wcs.wcs.aux.hgln_obs,
        'HGLT_OBS': aia171_test_map.wcs.wcs.aux.hglt_obs,
        'DSUN_OBS': aia171_test_map.wcs.wcs.aux.dsun_obs,
    }
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        return HighLevelWCSWrapper(SlicedLowLevelWCS(WCS(header=header), np.s_[0, :, :]))


@figure_test
def test_draw_extent_3d(aia171_test_map, wcs_3d_ln_lt_l_coupled):
    fig = Figure()
    ax = fig.add_subplot(projection=aia171_test_map)
    aia171_test_map.plot(axes=ax)
    drawing.extent(ax, wcs_3d_ln_lt_l_coupled)
    return fig



"""
Test Composite Map
"""
import matplotlib as mpl
import numpy as np
import pytest

import astropy.units as u
from astropy.tests.helper import assert_quantity_allclose

import sunpy.data.test
import sunpy.map
from sunpy.tests.helpers import figure_test

# Ignore missing metadata warnings
pytestmark = [pytest.mark.filterwarnings('ignore:Missing metadata for observer'),
              pytest.mark.filterwarnings(r'ignore:Unable to treat `\.meta` as a FITS header')]


@pytest.fixture
def composite_test_map(adjusted_test_maps):
    aia171_test_map, hmi_test_map = adjusted_test_maps
    return sunpy.map.Map(aia171_test_map, hmi_test_map, composite=True)


@pytest.fixture
def composite_test_map_prealigned(adjusted_test_maps):
    aia171_test_map, hmi_test_map = adjusted_test_maps
    hmi_test_map = hmi_test_map.reproject_to(aia171_test_map.wcs)
    return sunpy.map.Map(aia171_test_map, hmi_test_map, composite=True)


def test_type_of_arguments_composite_map(composite_test_map):
    with pytest.raises(ValueError, match="CompositeMap expects pre-constructed map objects.") as excinfo:
        sunpy.map.CompositeMap(23, composite=True)
    assert str(excinfo.value) == 'CompositeMap expects pre-constructed map objects.'


@figure_test
def test_autoalign_needed(composite_test_map):
    # The overplot will use pcolormesh if the WCSs do not match
    images = composite_test_map.plot()
    assert isinstance(images[0], mpl.image.AxesImage)
    assert isinstance(images[1], mpl.collections.QuadMesh)


@figure_test
def test_autoalign_not_needed(composite_test_map_prealigned):
    # The overplot will use imshow if the WCSs match
    images = composite_test_map_prealigned.plot()
    assert isinstance(images[0], mpl.image.AxesImage)
    assert isinstance(images[1], mpl.image.AxesImage)


@figure_test
def test_plot_composite_map(composite_test_map):
    composite_test_map.plot()


@figure_test
def test_plot_composite_map_contours(composite_test_map):
    composite_test_map.set_levels(1, np.arange(-75, 76, 25) << u.percent)
    composite_test_map.plot()


@figure_test
def test_plot_composite_map_linewidths(composite_test_map):
    composite_test_map.set_levels(1, np.arange(-75, 76, 25) << u.percent)
    composite_test_map.plot(linewidths=0.5)


@figure_test
def test_plot_composite_map_linestyles(composite_test_map):
    composite_test_map.set_levels(1, np.arange(-75, 76, 25) << u.percent)
    composite_test_map.plot(linestyles='--')


@figure_test
def test_plot_composite_map_colors(composite_test_map):
    composite_test_map.set_levels(1, np.arange(-75, 76, 25) << u.percent)
    composite_test_map.plot(colors='red')


def test_plot_composite_map_mplkwargs(composite_test_map):
    composite_test_map.set_levels(1, np.arange(-75, 76, 25) << u.percent)
    with pytest.raises(TypeError) as e:
        composite_test_map.plot(linestyles='--', unused_a=1, unused_b=2)
    assert 'plot() got unexpected keyword arguments' in str(e.value)
    assert 'unused_a' in str(e.value)
    assert 'unused_b' in str(e.value)
    assert 'linestyles' not in str(e.value)


def test_remove_composite_map(composite_test_map):
    composite_test_map.remove_map(0)
    with pytest.raises(IndexError):
        composite_test_map.get_map(1)


def test_get_composite_map(composite_test_map, aia171_test_map, hmi_test_map):
    assert composite_test_map.get_map(0) == aia171_test_map
    assert composite_test_map.get_map(1) == hmi_test_map


def test_get_alpha_composite_map(composite_test_map, aia171_test_map, hmi_test_map):
    assert composite_test_map.get_alpha() == [aia171_test_map.alpha, hmi_test_map.alpha]


def test_get_alpha_with_index_composite_map(composite_test_map, aia171_test_map, hmi_test_map):
    assert composite_test_map.get_alpha(0) == aia171_test_map.alpha
    assert composite_test_map.get_alpha(1) == hmi_test_map.alpha


def test_get_levels_composite_map(composite_test_map, aia171_test_map, hmi_test_map):
    assert composite_test_map.get_levels() == [aia171_test_map.levels, hmi_test_map.levels]


def test_get_levels_with_index_composite_map(composite_test_map, aia171_test_map, hmi_test_map):
    assert composite_test_map.get_levels(0) == aia171_test_map.levels
    assert composite_test_map.get_levels(1) == hmi_test_map.levels


@figure_test
def test_set_alpha_composite_map(composite_test_map):
    composite_test_map.set_alpha(1, 0.5)
    composite_test_map.plot()


@pytest.mark.parametrize(('index', 'alpha'), [(0, 5.0), (1, -3.0)])
def test_set_alpha_out_of_range_composite_map(composite_test_map, index, alpha):
    with pytest.raises(Exception, match="Alpha value must be between 0 and 1") as excinfo:
        composite_test_map.set_alpha(index, alpha)
    assert str(excinfo.value) == 'Alpha value must be between 0 and 1.'


def test_set_levels_percent(composite_test_map):
    numbers = np.arange(10, 100, 10)
    composite_test_map.set_levels(0, numbers)
    np.testing.assert_allclose(composite_test_map.get_levels(0), numbers)

    implicit_percentage = np.arange(10, 100, 10)
    composite_test_map.set_levels(0, implicit_percentage, percent=True)
    assert_quantity_allclose(composite_test_map.get_levels(0), implicit_percentage << u.percent)


@figure_test
def test_peek_composite_map(composite_test_map):
    composite_test_map.peek()



"""
This module implements a function to understand chaincodes from HELIO.
"""
import numpy as np

__all__ = ["Chaincode"]


class Chaincode(np.ndarray):
    """
    A tool to infer some information from chaincodes produced by HELIO Feature
    Catalogue or Heliophysics Events Knowledgebase.

    Parameters
    ----------
    origin : `numpy.ndarray`, `list`
        The 2 points of the origin of the chaincode.
    chaincode : `str`
        A list of the numbers (0-7) that indicate the path of the chaincode.
        0 moves horizontally to the left and the rest follows anticlockwise.
    xdelta : `float`, optional
        The scale to convert between pixels and flat coordinates. Defaults to 1.0.
    ydelta : `float`, optional
        The scale to convert between pixels and flat coordinates. Defaults to 1.0

    Returns
    -------
    `numpy.ndarray`
        An array containing all the x and y coordinates of the chaincode
        e.g., ``[[x0, x1, x2, ..., xn], [y0 ,y1, y2, ..., yn]]``.

    Examples
    --------
    >>> from sunpy.net.helio.chaincode import Chaincode
    >>> cc = Chaincode([-88, 812], "44464655567670006011212222324",
    ...     xdelta=2.629, ydelta=2.629)
    >>> fig = plt.figure()   # doctest: +SKIP
    >>> ax = fig.add_subplot(111)   # doctest: +SKIP
    >>> x,y = zip(cc.coordinates)   # doctest: +SKIP
    >>> ax.plot(x[0], y[0], 'go-')   # doctest: +SKIP
    >>> fig.show()   # doctest: +SKIP
    """
    def __new__(cls, origin, chaincode, **kwargs):
        if isinstance(origin, list):
            obj = np.asarray(origin).view(cls)
        elif isinstance(origin, np.ndarray):
            obj = origin.view(cls)
        else:
            raise TypeError('Invalid input')
        return obj

    def __init__(self, origin, chaincode, xdelta=1, ydelta=1):
        x_steps = [-1, -1, 0, 1, 1, 1, 0, -1]
        y_steps = [0, -1, -1, -1, 0, 1, 1, 1]
        self.coordinates = np.ndarray((2, len(chaincode) + 1))
        self.coordinates[:, 0] = origin
        if chaincode.isdigit():
            for index, step in enumerate(chaincode):
                self.coordinates[:, index + 1] = self.coordinates[:, index] + \
                    [[x_steps[int(step)] * xdelta,
                      y_steps[int(step)] * ydelta]]

    def matchend(self, end):
        return np.all(np.equal(self.coordinates[:, -1], np.asarray(end)))

    def matchany(self, coordinates, index):
        return np.all(
            np.allclose(self.coordinates[:, index], np.asarray(coordinates))
        )

    def boundingbox(self):
        """
        Extract the coordinates of the chaincode.

        Returns in the form of ``[[x0, x1], [y0, y1]]``.
        """
        bb = np.zeros((2, 2))
        bb[:, 0] = self.coordinates.min(1)
        bb[:, 1] = self.coordinates.max(1)
        return bb

    def area(self):
        raise NotImplementedError

    def length(self):
        raise NotImplementedError

    def sub_boundingbox(self, xedge=None, yedge=None):
        """
        Extract the x or y boundaries of the chaincode from a defined limits
        ``xedge`` or ``yedge``.

        Parameters
        ----------
        xedge : list, optional
            A list of two values to check.
        yedge : list, optional
            A list of two values to check.
        """
        if xedge is not None:
            edge = xedge
            IndexMask = 0
            IndexValue = 1
        elif yedge is not None:
            edge = yedge
            IndexMask = 1
            IndexValue = 0
        else:
            raise ValueError("Please input either `xedge` or `yedge`")
        mask = (self.coordinates[IndexMask, :] >= edge[0]) & (self.coordinates[IndexMask, :] <= edge[1])
        mx = np.ma.masked_array(self.coordinates[IndexValue, :], mask=(~mask))
        return [mx.min(), mx.max()]



"""
This module provides `sunpy.timeseries.GenericTimeSeries` which all other
`sunpy.timeseries.TimeSeries` classes inherit from.
"""
import copy
import html
import time
import textwrap
import webbrowser
from tempfile import NamedTemporaryFile
from collections import OrderedDict
from collections.abc import Iterable

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

import astropy
import astropy.units as u
from astropy.table import Column, Table
from astropy.time import Time
from astropy.visualization import hist

from sunpy import config
from sunpy.time import TimeRange
from sunpy.timeseries import TimeSeriesMetaData
from sunpy.util.datatype_factory_base import NoMatchError
from sunpy.util.exceptions import warn_user
from sunpy.util.metadata import MetaDict
from sunpy.util.util import _figure_to_base64
from sunpy.visualization import peek_show

# define and register a new unit, needed for RHESSI
det = u.def_unit('detector')
u.add_enabled_units([det])

TIME_FORMAT = config.get("general", "time_format")

__all__ = ["GenericTimeSeries"]


class GenericTimeSeries:
    """
    A generic time series object.

    Parameters
    ----------
    data : `~pandas.DataFrame` or `numpy.array`
        A `pandas.DataFrame` or `numpy.array` representing one or more fields as a function of time.
    meta : `~sunpy.timeseries.metadata.TimeSeriesMetaData`, optional
        The metadata giving details about the time series data/instrument.
        Defaults to `None`.
    units : `dict`, optional
        A mapping from column names in ``data`` to the physical units of that column.
        Defaults to `None`.

    Attributes
    ----------
    meta : `~sunpy.timeseries.metadata.TimeSeriesMetaData`
        The metadata giving details about the time series data/instrument.
    units : `dict`
        A mapping from column names in ``data`` to the physical units of that column.

    Examples
    --------
    >>> from sunpy.timeseries import TimeSeries
    >>> from sunpy.time import parse_time
    >>> from astropy.time import TimeDelta
    >>> import astropy.units as u
    >>> import numpy as np
    >>> import pandas as pd
    >>> times = parse_time("now") - TimeDelta(np.arange(24 * 60)*u.minute)
    >>> intensity = np.sin(np.arange(0, 12 * np.pi, step=(12 * np.pi) / (24 * 60)))
    >>> df = pd.DataFrame(intensity, index=times, columns=['intensity'])
    >>> header = {}
    >>> units = {'intensity': u.W/u.m**2}
    >>> ts = TimeSeries(df, header, units)
    >>> ts.peek()  # doctest: +SKIP

    References
    ----------
    * `Pandas Documentation <https://pandas.pydata.org/pandas-docs/stable/>`__
    """
    # Class attribute used to specify the source class of the TimeSeries.
    _source = None
    _registry = dict()

    # Title to show when .peek()ing
    _peek_title = ''

    def __init_subclass__(cls, **kwargs):
        """
        An __init_subclass__ hook initializes all of the subclasses of a given
        class.

        So for each subclass, it will call this block of code on import.
        This replicates some metaclass magic without the need to be
        aware of metaclasses. Here we use this to register each subclass
        in a dict that has the `is_datasource_for` attribute. This is
        then passed into the TimeSeries Factory so we can register them.
        """
        super().__init_subclass__(**kwargs)
        if hasattr(cls, 'is_datasource_for'):
            cls._registry[cls] = cls.is_datasource_for

    # kwargs are not used here but are passed in for sources.
    def __init__(self, data, meta=None, units=None, **kwargs):
        self._data = data
        tr = self.time_range
        # Check metadata input
        if meta is None:
            # No meta given, so default
            self.meta = TimeSeriesMetaData(MetaDict(), tr, self.columns)
        elif isinstance(meta, dict | OrderedDict | MetaDict):
            # Given the values for metadata (dict) and infer timerange and colnames from the data
            self.meta = TimeSeriesMetaData(meta, tr, self.columns)
        elif isinstance(meta, tuple):
            # Given the values all in a tuple
            self.meta = TimeSeriesMetaData(meta, tr, self.columns)
        else:
            # Should have a list of 3-tuples giving a complex metadata list.
            self.meta = meta

        if units is None:
            self.units = {}
        else:
            self.units = units

        for col in self.columns:
            if col not in self.units:
                warn_user(f'Unknown units for {col}')
                self.units[col] = u.dimensionless_unscaled

# #### Attribute definitions #### #

    @property
    def data(self):
        """
        A `pandas.DataFrame` representing one or more fields as a function of time.
        """
        warn_user("Using .data to access the dataframe is discouraged; "
                  "use .to_dataframe() instead.")
        return self._data

    @data.setter
    def data(self, d):
        self._data = d

    @property
    def source(self):
        """
        A string/object used to specify the source class of the TimeSeries.
        """
        return self._source

    @property
    def observatory(self):
        """
        A string/object used to specify the observatory for the TimeSeries.
        """
        return

    @property
    def columns(self):
        """
        A list of all the names of the columns in the data.
        """
        return list(self._data.columns.values)

    @property
    def time(self):
        """
        The timestamps of the data.
        """
        t = Time(self._data.index)
        # Set time format to enable plotting with astropy.visualisation.time_support()
        t.format = 'iso'
        return t

    @property
    def shape(self):
        """
        The shape of the data, a tuple (nrows, ncols).
        """
        return self._data.shape

    @property
    def time_range(self):
        """
        The start and end times of the TimeSeries as a `~sunpy.time.TimeRange`.
        """
        if len(self._data) > 0:
            return TimeRange(self._data.index.min(), self._data.index.max())
        else:
            return None

    @property
    def url(self):
        """
        URL to the mission website.
        """
        return self._url

# #### Data Access, Selection and Organisation Methods #### #

    def _text_summary(self):
        """
        Produces a table summary of the timeseries data to be passed to
        the _repr_html_ function.
        """
        obs = self.observatory
        if obs is None:
            try:
                obs = self.meta.metadata[0][2]["telescop"]
            except KeyError:
                obs = "Unknown"
        try:
            inst = self.meta.metadata[0][2]["instrume"]
        except KeyError:
            inst = "Unknown"
        try:
            link = f"""<a href={self.url} target="_blank">{inst}</a>"""
        except AttributeError:
            link = inst
        dat = self.to_dataframe()
        drange = dat.max() - dat.min()
        drange = drange.to_string(float_format="{:.2E}".format)
        drange = drange.replace("\n", "<br>")

        center = self.time_range.center.value.astype('datetime64[s]')
        center = str(center).replace("T", " ")
        resolution = round(self.time_range.seconds.value/self.shape[0], 3)
        resolution = str(resolution)+" s"

        channels = self.columns
        channels = "<br>".join(channels)

        uni = list(set(self.units.values()))
        uni = [x.unit if isinstance(x, u.quantity.Quantity) else x for x in uni]
        uni = ["dimensionless" if x == u.dimensionless_unscaled else x for x in uni]
        uni = "<br>".join(str(x) for x in uni)

        return textwrap.dedent(f"""\
                   SunPy TimeSeries
                   ----------------
                   Observatory:\t\t\t{obs}
                   Instrument:\t\t\t{link}
                   Channel(s):\t\t\t{channels}
                   Start Date:\t\t\t{dat.index.min().round('s')}
                   End Date:\t\t\t{dat.index.max().round('s')}
                   Center Date:\t\t\t{center}
                   Resolution:\t\t\t{resolution}
                   Samples per Channel:\t\t{self.shape[0]}
                   Data Range(s):\t\t{drange}
                   Units:\t\t\t{uni}""")

    def __str__(self):
        return f"{self._text_summary()}\n{self._data.__repr__()}"

    def __repr__(self):
        return f"{object.__repr__(self)}\n{self}"

    def _repr_html_(self):
        """
        Produces an HTML summary of the timeseries data with plots for use in
        Jupyter notebooks.
        """
        # Call _text_summary and reformat as an HTML table
        partial_html = (
            self._text_summary()[34:]
            .replace("\n", "</td></tr><tr><th>")
            .replace(":\t", "</th><td>")
        )
        text_to_table = (
            f"""\
            <table style='text-align:left'>
                <tr><th>{partial_html}</td></tr>
            </table>"""
        ).replace("\n", "")

        # Create the timeseries plots for each channel as a panel in one
        # figure. The color list below is passed to both timeseries and
        # histogram plotting methods for consistency.
        cols = ['b', 'g', 'r', 'c', 'm', 'y', 'tab:blue', 'tab:orange',
                'tab:red', 'tab:purple', 'tab:brown', 'tab:pink', 'tab:gray',
                'tab:green', 'tab:olive', 'tab:cyan', 'palegreen', 'pink'
                ]
        dat = self.to_dataframe()
        fig, axs = plt.subplots(
            nrows=len(self.columns),
            ncols=1,
            sharex=True,
            constrained_layout=True,
            figsize=(6, 10),
        )
        # If all channels have the same unit, then one shared y-axis
        # label is set. Otherwise, each subplot has its own yaxis label.
        for i in range(len(self.columns)):
            if len(self.columns) == 1:
                axs.plot(
                    dat.index,
                    dat[self.columns[i]].values,
                    color=cols[i%len(cols)],
                    label=self.columns[i],
                )
                if (dat[self.columns[i]].values < 0).any() is np.bool_(False):
                    axs.set_yscale("log")
                axs.legend(frameon=False, handlelength=0)
                axs.set_ylabel(self.units[self.columns[i]])
            else:
                axs[i].plot(
                    dat.index,
                    dat[self.columns[i]].values,
                    color=cols[i%len(cols)],
                    label=self.columns[i],
                )
                if (dat[self.columns[i]].values < 0).any() is np.bool_(False):
                    axs[i].set_yscale("log")
                axs[i].legend(frameon=False, handlelength=0)
                axs[i].set_ylabel(self.units[self.columns[i]])
        plt.xticks(rotation=30)
        spc = _figure_to_base64(fig)
        plt.close(fig)
        # Make histograms for each column of data. The histograms are
        # created using the Astropy hist method that uses Scott's rule
        # for bin sizes.
        hlist = []
        for i in range(len(dat.columns)):
            if set(np.isnan(dat[self.columns[i]].values)) != {True}:
                fig = plt.figure(figsize=(5, 3), constrained_layout=True)
                hist(
                    dat[self.columns[i]].values[~np.isnan(dat[self.columns[i]].values)],
                    log=True,
                    bins="scott",
                    color=cols[i%len(cols)],
                )
                plt.title(self.columns[i] + " [click for other channels]")
                plt.xlabel(self.units[self.columns[i]])
                plt.ylabel("# of occurrences")
                hlist.append(_figure_to_base64(fig))
                plt.close(fig)

        # This loop creates a formatted list of base64 images that is passed
        # directly into the JS script below, so all images are written into
        # the html page when it is created (allows for an arbitrary number of
        # histograms to be rotated through onclick).
        hlist2 = []
        for i in range(len(hlist)):
            hlist2.append(f"data:image/png;base64,{hlist[i]}")

        # The code below creates unique names to be passed to the JS script
        # in the html code. Otherwise, multiple timeseries summaries will
        # conflict in a single notebook.
        source = str(self.source) + str(time.perf_counter_ns())

        return textwrap.dedent(f"""\
            <pre>{html.escape(object.__repr__(self))}</pre>
            <script type="text/javascript">
            function ImageChange(images) {{
                this.images = images;
                this.i = 0;
                this.next = function(img) {{
                    this.i++;
                    if (this.i == images.length)
                    this.i = 0;
                    img.src = images[this.i];
                }}
            }}
            var {source} = new ImageChange({hlist2});
            </script>
            <table>
                <tr>
                    <td style='width:40%'>{text_to_table}</td>
                    <td rowspan=3>
                        <img src='data:image/png;base64,{spc}'/>
                    </td>
                </tr>
                <tr>
                </tr>
                <tr>
                    <td>
                    <img src="{hlist2[0]}" alt="Click here for histograms"
                         onclick="{source}.next(this)"/>
                    </td>
                </tr>
            </table>""")

    def quicklook(self):
        """
        Display a quicklook summary of the Timeseries instance in the default
        webbrowser.

        Example
        -------
        >>> from sunpy.timeseries import TimeSeries
        >>> import sunpy.data.sample  # doctest: +REMOTE_DATA
        >>> goes_lc = TimeSeries(sunpy.data.sample.GOES_XRS_TIMESERIES)  # doctest: +REMOTE_DATA
        >>> goes_lc.quicklook()  # doctest: +SKIP
        """
        with NamedTemporaryFile(
            "w", delete=False, prefix="sunpy.timeseries.", suffix=".html"
        ) as f:
            url = "file://" + f.name
            f.write(textwrap.dedent(f"""\
                <html>
                    <title>Quicklook summary for {html.escape(object.__repr__(self))}</title>
                    <body>{self._repr_html_()}</body>
                </html>""")
                    )
        webbrowser.open_new_tab(url)

    def quantity(self, colname, **kwargs):
        """
        Return a `~astropy.units.quantity.Quantity` for the given column.

        Parameters
        ----------
        colname : `str`
            The heading of the column you want to output.

        Returns
        -------
        `~astropy.units.quantity.Quantity`
        """
        values = self._data[colname].values
        unit = self.units[colname]
        return u.Quantity(values, unit)

    def add_column(self, colname, quantity, unit=False, overwrite=True, **kwargs):
        """
        Return a new `~sunpy.timeseries.TimeSeries` with the given column added
        or updated.

        Parameters
        ----------
        colname : `str`
            The heading of the column you want output.
        quantity : `~astropy.units.quantity.Quantity` or `~numpy.ndarray`
            The values to be placed within the column.
            If updating values only then a numpy array is permitted.
        overwrite : `bool`, optional
            Defaults to `True`, allowing the method to overwrite a column already present in the `~sunpy.timeseries.TimeSeries`.

        Returns
        -------
        `sunpy.timeseries.TimeSeries`
            A new `~sunpy.timeseries.TimeSeries`.
        """
        # Get the expected units from the quantity if required
        if not unit and isinstance(quantity, astropy.units.quantity.Quantity):
            unit = quantity.unit
        elif not unit:
            unit = u.dimensionless_unscaled

        # Make a copy of all the TimeSeries components.
        data = copy.copy(self._data)
        meta = TimeSeriesMetaData(copy.copy(self.meta.metadata))
        units = copy.copy(self.units)

        # Add the unit to the units dictionary if already there.
        if colname not in self.columns:
            units[colname] = unit

        # Convert the given quantity into values for given units if necessary.
        values = quantity
        if isinstance(values, astropy.units.quantity.Quantity) and overwrite:
            values = values.to(units[colname]).value

        # Update or add the data.
        if colname not in self.columns or overwrite:
            data[colname] = values

        # Return a new TimeSeries with the given updated/added column.
        return self.__class__(data, meta, units)

    def remove_column(self, colname):
        """
        Remove a column.

        Parameters
        ----------
        colname : str
            The heading of the column to remove.

        Returns
        -------
        `sunpy.timeseries.TimeSeries`
            A new `~sunpy.timeseries.TimeSeries`.
        """
        if colname not in self.columns:
            raise ValueError(f'Given column name ({colname}) not in list of columns {self.columns}')
        data = self._data.drop(colname, axis='columns')
        units = self.units.copy()
        units.pop(colname)
        return self.__class__(data, self.meta, units)

    def sort_index(self, **kwargs):
        """
        Returns a sorted version of a `~sunpy.timeseries.TimeSeries`. Generally
        this shouldn't be necessary as most `~sunpy.timeseries.TimeSeries`
        operations sort the data anyway to ensure consistent behavior when
        truncating.

        Returns
        -------
        `~sunpy.timeseries.TimeSeries`
            A new `~sunpy.timeseries.TimeSeries` in ascending chronological order.
        """
        return GenericTimeSeries(self._data.sort_index(**kwargs),
                                 TimeSeriesMetaData(copy.copy(self.meta.metadata)),
                                 copy.copy(self.units))

    def truncate(self, a, b=None, int=None):
        """
        Returns a truncated version of the TimeSeries object.

        Parameters
        ----------
        a : `sunpy.time.TimeRange`, `str`, `int`
            Either a time range to truncate to, or a start time in some format recognized by pandas, or a index integer.
        b : `str` or `int`, optional
            If specified, the end time of the time range in some format recognized by pandas, or a index integer.
            Defaults to `None`.
        int : `int`, optional
            If specified, the integer indicating the slicing intervals.
            Defaults to `None`.

        Returns
        -------
        `~sunpy.timeseries.TimeSeries`
            A new `~sunpy.timeseries.TimeSeries` with only the selected times.
        """
        # Evaluate inputs
        # If given strings, then use to create a sunpy.time.timerange.TimeRange
        # for the SunPy text date parser.
        if isinstance(a, str) and isinstance(b, str):
            a = TimeRange(a, b)
        if isinstance(a, TimeRange):
            # If we have a TimeRange, extract the values
            start = a.start.datetime
            end = a.end.datetime
        else:
            # Otherwise we already have the values
            start = a
            end = b

        # If an interval integer was given then use in truncation.
        truncated_data = self._data.sort_index()[start:end:int]

        # Truncate the metadata
        # Check there is data still
        truncated_meta = TimeSeriesMetaData([])
        if len(truncated_data) > 0:
            tr = TimeRange(truncated_data.index.min(), truncated_data.index.max())
            truncated_meta = TimeSeriesMetaData(copy.deepcopy(self.meta.metadata))
            truncated_meta._truncate(tr)

        # Build similar TimeSeries object and sanatise metadata and units.
        object = self.__class__(truncated_data.sort_index(), truncated_meta, copy.copy(self.units))
        object._sanitize_metadata()
        return object

    def extract(self, column_name):
        """
        Returns a new time series with the chosen column.

        Parameters
        ----------
        column_name : `str`
            A valid column name.

        Returns
        -------
        `~sunpy.timeseries.TimeSeries`
            A new `~sunpy.timeseries.TimeSeries` with only the selected column.
        """
        # TODO: allow the extract function to pick more than one column
        # TODO: Fix this?
        # if isinstance(self, pandas.Series):
        #    return self
        # else:
        #    return GenericTimeSeries(self._data[column_name], TimeSeriesMetaData(self.meta.metadata.copy()))

        # Extract column and remove empty rows
        data = self._data[[column_name]].dropna()
        units = {column_name: self.units[column_name]}

        # Build generic TimeSeries object and sanatise metadata and units.
        object = GenericTimeSeries(data.sort_index(),
                                   TimeSeriesMetaData(copy.copy(self.meta.metadata)),
                                   units)
        object._sanitize_metadata()
        return object

    def concatenate(self, others, same_source=False, **kwargs):
        """
        Concatenate with another `~sunpy.timeseries.TimeSeries` or an iterable containing multiple
        `~sunpy.timeseries.TimeSeries`. This function will check and remove any duplicate times.
        It will keep the column values from the original timeseries to which the new time
        series is being added.

        Parameters
        ----------
        others : `~sunpy.timeseries.TimeSeries` or `collections.abc.Iterable`
            Another `~sunpy.timeseries.TimeSeries` or an iterable containing multiple
            `~sunpy.timeseries.TimeSeries`.
        same_source : `bool`, optional
            Set to `True` to check if the sources of the time series match. Defaults to `False`.

        Returns
        -------
        `~sunpy.timeseries.TimeSeries`
            A new `~sunpy.timeseries.TimeSeries`.

        Notes
        -----
        Extra keywords are passed to `pandas.concat`.

        Examples
        --------
        A single `~sunpy.timeseries.TimeSeries` or an `collections.abc.Iterable` containing multiple
        `~sunpy.timeseries.TimeSeries` can be passed to concatenate.

        >>> timeseries_1.concatenate(timeseries_2) # doctest: +SKIP
        >>> timeseries_1.concatenate([timeseries_2, timeseries_3]) # doctest: +SKIP

        Set ``same_source`` to `True` if the sources of the time series are the same.

        >>> timeseries_1.concatenate([timeseries_2, timeseries_3], same_source=True) # doctest: +SKIP
        """
        # Check to see if nothing needs to be done in case the same TimeSeries is provided.
        if self == others:
            return self
        elif isinstance(others, Iterable):
            if len(others) == 1 and self == next(iter(others)):
                return self

        # Check the sources match if specified.
        if (
            same_source
            and isinstance(others, Iterable)
            and not all(isinstance(series, self.__class__) for series in others)
        ):
            raise TypeError("TimeSeries classes must match if 'same_source' is specified.")
        elif (
            same_source
            and not isinstance(others, Iterable)
            and not isinstance(others, self.__class__)
        ):
            raise TypeError("TimeSeries classes must match if 'same_source' is specified.")

        # If an iterable is not provided, it must be a TimeSeries object, so wrap it in a list.
        if not isinstance(others, Iterable):
            others = [others]

        # Concatenate the metadata and data.
        kwargs["sort"] = kwargs.pop("sort", False)
        meta = self.meta.concatenate([series.meta for series in others])
        data = pd.concat(
            [self._data.copy(), *list(series.to_dataframe() for series in others)], **kwargs
        )

        # Add all the new units to the dictionary.
        units = OrderedDict()
        units.update(self.units)
        units.update(
            {k: v for unit in list(series.units for series in others) for k, v in unit.items()}
        )
        units = {k: v for k, v in units.items() if k in data.columns}

        # If sources match then build similar TimeSeries.
        if all(self.__class__ == series.__class__ for series in others):
            object = self.__class__(data.sort_index(), meta, units)
        else:
            # Build generic time series if the sources don't match.
            object = GenericTimeSeries(data.sort_index(), meta, units)

        # Sanatise metadata and units
        object._sanitize_metadata()
        return object

# #### Plotting Methods #### #

    def plot(self, axes=None, columns=None, **plot_args):
        """
        Plot a plot of the `~sunpy.timeseries.TimeSeries`.

        Parameters
        ----------
        axes : `~matplotlib.axes.Axes`, optional
            If provided the image will be plotted on the given axes.
            Defaults to `None`, so the current axes will be used.
        columns : list[str], optional
            If provided, only plot the specified columns.
        **plot_args : `dict`, optional
            Additional plot keyword arguments that are handed to
            :meth:`pandas.DataFrame.plot`.

        Returns
        -------
        `~matplotlib.axes.Axes`
            The plot axes.
        """
        axes, columns = self._setup_axes_columns(axes, columns)

        axes = self._data[columns].plot(ax=axes, **plot_args)

        units = set([self.units[col] for col in columns])
        if len(units) == 1:
            # If units of all columns being plotted are the same, add a unit
            # label to the y-axis.
            unit = u.Unit(list(units)[0])
            axes.set_ylabel(unit.to_string())

        self._setup_x_axis(axes)
        return axes

    def _setup_axes_columns(self, axes, columns, *, subplots=False):
        """
        Validate data for plotting, and get default axes/columns if not passed
        by the user.
        """
        import matplotlib.pyplot as plt

        self._validate_data_for_plotting()
        if columns is None:
            columns = self.columns
        if axes is None:
            if not subplots:
                axes = plt.gca()
            else:
                axes = plt.gcf().subplots(ncols=1, nrows=len(columns), sharex=True)

        return axes, columns

    @staticmethod
    def _setup_x_axis(ax):
        """
        Shared code to set x-axis properties.
        """
        import matplotlib.dates as mdates
        if isinstance(ax, np.ndarray):
            ax = ax[-1]

        locator = ax.xaxis.get_major_locator()
        formatter = ax.xaxis.get_major_formatter()
        if isinstance(formatter, mdates.AutoDateFormatter):
            # Override Matplotlib default date formatter (concise one is better)
            # but don't override any formatters pandas might have set
            ax.xaxis.set_major_formatter(mdates.ConciseDateFormatter(locator))

    @peek_show
    def peek(self, *, columns=None, title=None, **kwargs):
        """
        Displays a graphical overview of the data in this object for user evaluation.
        For the creation of plots, users should instead use the
        `~sunpy.timeseries.GenericTimeSeries.plot` method and Matplotlib's pyplot framework.

        Parameters
        ----------
        columns : list[str], optional
            If provided, only plot the specified columns.
        title : str, optional
            If provided, set the plot title. Custom timeseries sources may set
            a default value for this.
        **kwargs : `dict`
            Any additional plot arguments that should be used when plotting.
        """
        import matplotlib.pyplot as plt

        # Now make the plot
        figure = plt.figure()
        self.plot(columns=columns, **kwargs)
        title = title or self._peek_title
        figure.suptitle(title)

        return figure

    def _validate_data_for_plotting(self):
        """
        Raises an exception if the `~sunpy.timeseries.TimeSeries` is invalid
        for plotting.

        This should be added into all `~sunpy.timeseries.TimeSeries`
        peek methods.
        """
        # Check we have a valid TS
        if len(self._data) == 0:
            raise ValueError("The timeseries can't be plotted as it has no data present. "
                             "(len(self._data) == 0)")

# #### Miscellaneous #### #

    def _sanitize_metadata(self, **kwargs):
        """
        Sanitizes the `~sunpy.timeseries.TimeSeriesMetaData`  used to store the
        metadata.

        Primarily this method will:

        * Remove entries outside of the dates or truncate if the metadata overflows past the data.
        * Remove column references in the metadata that don't match to a column in the data.
        * Remove metadata entries that have no columns matching the data.
        """
        # Truncate the metadata
        self.meta._truncate(self.time_range)

        # Remove non-existent columns
        redundant_cols = list(set(self.meta.columns) - set(self.columns))
        self.meta._remove_columns(redundant_cols)

# #### Export/Output Methods #### #

    def to_table(self, **kwargs):
        """
        Return an `astropy.table.Table` of the given
        `~sunpy.timeseries.TimeSeries`.

        Returns
        -------
        `~astropy.table.Table`
            A new `astropy.table.Table` containing the data from the `~sunpy.timeseries.TimeSeries`.
            The table will include units where relevant.
        """
        # TODO: Table.from_pandas(df) doesn't include the index column. Add request?
        # Get data columns
        table = Table.from_pandas(self._data)

        # Get index column and add to table.
        index_col = Column(self._data.index.values, name='date')
        table.add_column(index_col, index=0)

        # Add in units.
        for key in self.units:
            table[key].unit = self.units[key]

        # Output the table
        return table

    def to_dataframe(self, **kwargs):
        """
        Return a `~pandas.DataFrame` of the given
        `~sunpy.timeseries.TimeSeries`.

        Returns
        -------
        `~pandas.DataFrame`
        """
        return self._data

    def to_array(self, **kwargs):
        """
        Return a `numpy.array` of the given `~sunpy.timeseries.TimeSeries`.

        Parameters
        ----------
        **kwargs : `dict`
            All keyword arguments are passed to ``pandas.DataFrame.to_numpy``.

        Returns
        -------
        `~numpy.ndarray`
            If the data is heterogeneous and contains booleans or objects, the result will be of ``dtype=object``.
        """
        if hasattr(self._data, "to_numpy"):
            return self._data.to_numpy(**kwargs)
        else:
            return self._data.values

    def __eq__(self, other):
        """
        Check two `~sunpy.timeseries.TimeSeries` are the same, they have
        matching type, data, metadata and units entries.

        Parameters
        ----------
        other : `~sunpy.timeseries.TimeSeries`
            The second `~sunpy.timeseries.TimeSeries` to compare with.

        Returns
        -------
        `bool`
        """
        match = True
        if isinstance(other, type(self)):
            if ((not self._data.equals(other.to_dataframe())) or
                    (self.meta != other.meta) or
                    (self.units != other.units)):
                match = False
        else:
            match = False
        return match

    def __ne__(self, other):
        """
        Check two `~sunpy.timeseries.TimeSeries` are not the same, they don't
        have matching type, data, metadata and/or units entries.

        Parameters
        ----------
        other : `~sunpy.timeseries.TimeSeries`
            The second `~sunpy.timeseries.TimeSeries` to compare with.

        Returns
        -------
        `bool`
        """
        return not self == other

    @classmethod
    def _parse_file(cls, filepath):
        """
        Parses a file - to be implemented in any subclass that may use files.

        Parameters
        ----------
        filepath : `str`
            The path to the file you want to parse.
        """
        raise NoMatchError(f'Could not find any timeseries sources to parse {filepath}')



import astropy.units as u

from sunpy.map.mapbase import GenericMap, SpatialPair

__all__ = ['RHESSIMap']


class RHESSIMap(GenericMap):
    """
    RHESSI Image Map.

    The RHESSI mission consists of a single spin-stabilized
    spacecraft in a low-altitude orbit inclined 38 degrees to
    the Earth's equator. The only instrument on board is an
    Germaniun imaging spectrometer with the ability to obtain high
    fidelity solar images in X rays (down to 3 keV) to gamma rays (1 MeV).

    RHESSI provides an angular resolution of 2 arcseconds at
    X-ray energies below ~40 keV, 7 arcseconds to 400 keV,
    and 36 arcseconds for gamma-ray lines and continuum above 1 MeV.

    RHESSI was launched on 5 February 2002.

    References
    ----------
    * RHESSI Homepage `<https://hesperia.gsfc.nasa.gov/rhessi3/index.html>`__
    * :cite:t:`lin_reuven_2002`

    .. warning::

        Cannot read fits files containing more than one image.

    Examples
    --------
    .. plot::
        :include-source:
        :context: reset

        import sunpy.map
        from sunpy.data.sample import RHESSI_IMAGE
        rhessi_map = sunpy.map.Map(RHESSI_IMAGE)
        rhessi_map.plot()
    """

    def __init__(self, data, header, **kwargs):
        super().__init__(data, header, **kwargs)
        self._nickname = self.detector
        self.plot_settings['cmap'] = 'rhessi'

    def _get_cmap_name(self):
        return "rhessi"

    @property
    def _timesys(self):
        """
        RHESSI maps can incorrectly use the TIMESYS keyword for the reference
        time. If this is the case, returns the FITS default UTC.
        """
        if ('TIMESYS' in self.meta and
                self.meta['keycomments']['TIMESYS'] == 'Reference Time'):
            return 'UTC'
        else:
            return super()._timesys

    def _rotation_matrix_from_crota(self):
        """
        RHESSI maps can have their rotation in CROTA.
        """
        return super()._rotation_matrix_from_crota(crota_key='CROTA')

    @property
    def spatial_units(self):
        """
        If CTYPE{1 or 2} are equal to 'arcsec', assumes that CTYPE{1 or 2}
        respectively are intended to be 'arcsec'.
        """
        units = [self.meta.get('cunit1', None), self.meta.get('cunit2', None)]
        if self.meta['ctype1'] == 'arcsec':
            units[0] = 'arcsec'
        if self.meta['ctype2'] == 'arcsec':
            units[1] = 'arcsec'
        units = [None if unit is None else u.Unit(unit.lower()) for unit in units]
        return SpatialPair(units[0], units[1])

    @property
    def coordinate_system(self):
        """
        If CTYPE{1 or 2} are equal to 'arcsec', assumes that CTYPE{1 or 2}
        respectively are intended to be 'HPLN-TAN' or 'HPLT-TAN'.
        """
        ctype1, ctype2 = self.meta['ctype1'], self.meta['ctype2']
        if ctype1 == 'arcsec':
            ctype1 = 'HPLN-TAN'
        if ctype2 == 'arcsec':
            ctype2 = 'HPLT-TAN'
        return SpatialPair(ctype1, ctype2)

    @property
    def waveunit(self):
        """
        If the WAVEUNIT FITS keyword is not present, defaults to keV.
        """
        unit = self.meta.get("waveunit", 'keV')
        return u.Unit(unit)

    @property
    def wavelength(self):
        return u.Quantity([self.meta['energy_l'], self.meta['energy_h']],
                          unit=self.waveunit)

    @property
    def detector(self):
        return self.meta['telescop']

    @classmethod
    def is_datasource_for(cls, data, header, **kwargs):
        """Determines if header corresponds to an RHESSI image"""
        return header.get('instrume') == 'RHESSI'



"""
==================================
Reprojecting with Automatic Extent
==================================

This example demonstrates how you can automatically determine the extent
needed for the output of a reprojection instead of specifying it ahead of time.
"""
# sphinx_gallery_thumbnail_number = 3

import matplotlib.pyplot as plt

import astropy.units as u
from astropy.coordinates import SkyCoord

import sunpy.map
from sunpy.coordinates import SphericalScreen
from sunpy.data.sample import AIA_171_IMAGE

######################################################################
# Let's build off of the final reprojection made in the example
# :ref:`sphx_glr_generated_gallery_map_transformations_reprojection_spherical_screen.py`.

aia_map = sunpy.map.Map(AIA_171_IMAGE)
aia_map.plot_settings['norm'].vmin = 0
aia_map.plot_settings['norm'].vmax = 10000

new_observer = SkyCoord(70*u.deg, 20*u.deg, 1*u.AU, obstime=aia_map.date,
                        frame='heliographic_stonyhurst')

out_shape = aia_map.data.shape

out_ref_coord = SkyCoord(0*u.arcsec, 0*u.arcsec, obstime=new_observer.obstime,
                         frame='helioprojective', observer=new_observer,
                         rsun=aia_map.coordinate_frame.rsun)
out_header = sunpy.map.make_fitswcs_header(
    out_shape,
    out_ref_coord,
    scale=u.Quantity(aia_map.scale),
    instrument=aia_map.instrument,
    wavelength=aia_map.wavelength
)

with SphericalScreen(aia_map.observer_coordinate, only_off_disk=True):
    outmap_screen_off_disk = aia_map.reproject_to(out_header)

fig = plt.figure()
ax = fig.add_subplot(projection=outmap_screen_off_disk)
outmap_screen_off_disk.plot(axes=ax)

######################################################################
# Note that some of the data is clipped at the top and bottom because
# ``out_shape`` was not defined large enough in the vertical
# direction. On the other hand, there is too much empty space to the
# left and right because ``out_shape`` was defined too large in the
# horizontal direction, which actually results in unnecessary
# computation time. That is also why one should avoid specifying
# a conservatively large ``out_shape``.
#
# Instead, we can enable the automatic determination of extent by
# :meth:`~sunpy.map.GenericMap.reproject_to` via the keyword
# ``auto_extent``.  Let's first try ``auto_extent='edges'``, which
# is often sufficient for many situations. This setting will
# precompute the reprojection of just the edges of the map to see
# what the output extent should be.

with SphericalScreen(aia_map.observer_coordinate, only_off_disk=True):
    outmap_edges = aia_map.reproject_to(out_header, auto_extent='edges')

fig = plt.figure()
ax = fig.add_subplot(projection=outmap_edges)
outmap_edges.plot(axes=ax)

######################################################################
# In this case, ``auto_extent='edges'`` misses the fact that the Sun
# in the middle part of the map actually extends to the left of one of
# edges in the reprojection. Thus, instead, we need to specify
# ``auto_extent='all'`` to precompute the reprojection of all pixels
# in the map.  This setting takes more time, but will ensure that no
# part of the reprojected map is clipped.

with SphericalScreen(aia_map.observer_coordinate, only_off_disk=True):
    outmap_all = aia_map.reproject_to(out_header, auto_extent='all')

fig = plt.figure()
ax = fig.add_subplot(projection=outmap_all)
outmap_all.plot(axes=ax)

plt.show()




import numpy as np
import pytest

from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.table import MaskedColumn, Table
from astropy.time import Time

from sunpy.coordinates import Helioprojective, get_earth
from sunpy.net import Fido, attr, attrs, hek
from sunpy.net.hek.utils import _get_coord_attributes, _get_unit_attributes


@pytest.fixture
def foostrwrap(request):
    return hek.attrs._StringParamAttrWrapper("foo")


@pytest.fixture(scope="session")
def hek_result():
    startTime = '2011/08/09 07:23:56'
    endTime = '2011/08/09 12:40:29'
    eventType = 'FL'
    hekTime = attrs.Time(startTime, endTime)
    hekEvent = attrs.hek.EventType(eventType)
    h = hek.HEKClient()
    return h.search(hekTime, hekEvent)


def test_eventtype_collide():
    with pytest.raises(TypeError):
        attrs.hek.AR & attrs.hek.CE
    with pytest.raises(TypeError):
        (attrs.hek.AR & attrs.Time((2011, 1, 1),
                                   (2011, 1, 2))) & attrs.hek.CE
    with pytest.raises(TypeError):
        (attrs.hek.AR | attrs.Time((2011, 1, 1),
                                   (2011, 1, 2))) & attrs.hek.CE


def test_eventtype_or():
    assert (attrs.hek.AR | attrs.hek.CE).item == "ar,ce"


def test_HEKAttr():
    res = hek.attrs.walker.create(hek.attrs.HEKAttr("foo", "=", "bar"), {})
    assert len(res) == 1
    assert res[0] == {'value0': 'bar', 'operator0': '=', 'param0': 'foo'}


def test_stringwrapper_eq(foostrwrap):
    res = hek.attrs.walker.create(foostrwrap == "bar", {})
    assert len(res) == 1
    assert res[0] == {'value0': 'bar', 'operator0': '=', 'param0': 'foo'}


def test_stringwrapper_lt(foostrwrap):
    res = hek.attrs.walker.create(foostrwrap < "bar", {})
    assert len(res) == 1
    assert res[0] == {'value0': 'bar', 'operator0': '<', 'param0': 'foo'}


def test_stringwrapper_gt(foostrwrap):
    res = hek.attrs.walker.create(foostrwrap > "bar", {})
    assert len(res) == 1
    assert res[0] == {'value0': 'bar', 'operator0': '>', 'param0': 'foo'}


def test_stringwrapper_le(foostrwrap):
    res = hek.attrs.walker.create(foostrwrap <= "bar", {})
    assert len(res) == 1
    assert res[0] == {'value0': 'bar', 'operator0': '<=', 'param0': 'foo'}


def test_stringwrapper_ge(foostrwrap):
    res = hek.attrs.walker.create(foostrwrap >= "bar", {})
    assert len(res) == 1
    assert res[0] == {'value0': 'bar', 'operator0': '>=', 'param0': 'foo'}


def test_stringwrapper_ne(foostrwrap):
    res = hek.attrs.walker.create(foostrwrap != "bar", {})
    assert len(res) == 1
    assert res[0] == {'value0': 'bar', 'operator0': '!=', 'param0': 'foo'}


def test_stringwrapper_like(foostrwrap):
    res = hek.attrs.walker.create(foostrwrap.like("bar"), {})
    assert len(res) == 1
    assert res[0] == {'value0': 'bar', 'operator0': 'like', 'param0': 'foo'}


def test_err_dummyattr_create():
    with pytest.raises(TypeError):
        hek.attrs.walker.create(attr.DummyAttr(), {})


def test_err_dummyattr_apply():
    with pytest.raises(TypeError):
        hek.attrs.walker.apply(attr.DummyAttr(), {})


@pytest.mark.remote_data
def test_hek_client(hek_result):
    assert isinstance(hek_result, hek.hek.HEKTable)


@pytest.mark.remote_data
def test_hek_empty_search_result():
    startTime = '1985-05-04 00:00:00'
    endTime = '1985-05-04 00:00:00'
    eventType = 'FL'

    hekTime = attrs.Time(startTime, endTime)
    hekEvent = attrs.hek.EventType(eventType)

    h = hek.HEKClient()
    hek_query = h.search(hekTime, hekEvent)
    assert isinstance(hek_query, hek.hek.HEKTable)
    assert len(hek_query) == 0


@pytest.mark.remote_data
def test_get_voevent(hek_result):
    ve = hek_result[0].get_voevent()
    assert len(ve['voe:VOEvent']) == 7


@pytest.mark.remote_data
def test_hek_time_col(hek_result):
    assert isinstance(hek_result[0]['event_starttime'], Time)
    assert isinstance(hek_result[0]['event_endtime'], Time)


@pytest.mark.remote_data
def test_vso_time(hek_result):
    ve = hek_result[0].vso_time
    assert isinstance(ve, attrs.Time)


@pytest.mark.remote_data
def test_vso_instrument(hek_result):
    vc = hek_result[1].vso_instrument
    assert isinstance(vc, attrs.Instrument)


@pytest.mark.remote_data
def test_HEKRow_get(hek_result):
    assert hek_result[0]['event_peaktime'] == hek_result[0].get('event_peaktime')
    assert hek_result[0].get('') is None


@pytest.mark.remote_data
def test_mixed_results_get():
    # To check that the following bug is fixed:
    # https://github.com/sunpy/sunpy/issues/3238
    client = hek.HEKClient()
    result = client.search(attrs.Time('2013/02/01 00:00:00', '2013/02/01 23:30:00'),
                           attrs.hek.FRM.Name == 'SPoCA')
    assert isinstance(result, hek.hek.HEKTable)
    assert len(result) == 89
    # We do not check the full timestamp as the last 8 digits change as data is reprocessed.
    assert result[0]["SOL_standard"].startswith("SOL2013-01-31T20:13:31")


@pytest.mark.remote_data
def test_mixed_results_get_2():
    # To check that the following bug is fixed:
    # # https://github.com/sunpy/sunpy/issues/3898
    client = hek.HEKClient()
    result = client.search(attrs.Time('2011/08/09 07:23:56', '2011/08/09 12:40:29'),
                           attrs.hek.EventType("FL"))
    assert isinstance(result, hek.hek.HEKTable)
    assert len(result) == 19
    # We do not check the full timestamp as the last 8 digits change as data is reprocessed.
    assert result[0]["SOL_standard"].startswith("SOL2011-08-08T01:30:04")


@pytest.mark.remote_data
def test_mixed_results_get_angstrom():
    # To check that the following bug is fixed:
    # https://github.com/sunpy/sunpy/issues/4087
    client = hek.HEKClient()
    tstart = '2014/10/24 20:50'
    tend = '2014/10/25 00:14'
    event_type = 'FL'
    result = client.search(attrs.Time(tstart, tend), attrs.hek.EventType(event_type))
    assert len(result) == 13
    assert result[0]["SOL_standard"] == 'SOL2014-10-24T20:53:46L247C106'


@pytest.mark.remote_data
def test_query_multiple_operators():
    event_type = "FL"
    tstart = "2013/10/28"
    tend = "2013/10/29"
    client = hek.HEKClient()
    results = client.search(attrs.Time(tstart, tend),
                            attrs.hek.EventType(event_type),
                            attrs.hek.FL.GOESCls > "M1.0",
                            attrs.hek.OBS.Observatory == "GOES")
    assert len(results) == 7


@pytest.mark.remote_data
def test_astropy_unit_parsing(coronal_hole_search_result):
    hek_attributes = _get_unit_attributes() + _get_coord_attributes()
    attributes_with_unit = [attribute for attribute in hek_attributes if attribute.get("unit_prop", False)]
    for attribute in attributes_with_unit:
        if attribute["name"] in coronal_hole_search_result.colnames and not attribute.get("is_chaincode", False):
            assert isinstance(coronal_hole_search_result[attribute['name']], u.Quantity)


@pytest.mark.remote_data
def test_chaincode_parsing(coronal_hole_search_result):
    chaincode_attributes = [attribute for attribute in _get_coord_attributes() if attribute.get("is_chaincode", False)]
    for attribute in chaincode_attributes:
        if attribute["name"] in coronal_hole_search_result.colnames:
            if isinstance(coronal_hole_search_result[attribute['name']], MaskedColumn):
                assert all([isinstance(r, SkyCoord) for r in coronal_hole_search_result[attribute['name']]])
            else:
                assert isinstance(coronal_hole_search_result[attribute['name']], SkyCoord)


@pytest.mark.remote_data
def test_missing_times():
    # Check for https://github.com/sunpy/sunpy/pull/7627#issuecomment-2113451964
    client = hek.HEKClient()
    results = client.search(attrs.Time('2024-05-10', '2024-05-12'), attrs.hek.AR.NOAANum == 13664)
    assert isinstance(results["event_peaktime"][0], np.ma.core.MaskedConstant)
    assert results["event_peaktime"][3].isot == "2024-05-10T00:13:00.000"


@pytest.fixture(scope="session")
def coronal_hole_search_result():
    client = hek.HEKClient()
    result = client.search(attrs.Time('2011/08/09 07:23:56', '2011/08/09 12:40:29'),
                           attrs.hek.EventType('CH'))
    return result


@pytest.mark.remote_data
def test_compare_event_coords(coronal_hole_search_result):
    event_coord = SkyCoord(-2.91584*u.arcsec,
                           940.667*u.arcsec,
                           frame=Helioprojective(observer=get_earth('2011-08-09 06:00:08.000')))
    assert coronal_hole_search_result['event_coord'][0] == event_coord


@pytest.mark.remote_data
def test_obs_meanwavel(coronal_hole_search_result):
    assert coronal_hole_search_result['obs_meanwavel'][0] == 193.0*u.angstrom


@pytest.fixture
def flare_search():
    return attrs.Time('2011/08/09 07:23:56', '2011/08/09 12:40:29'), attrs.hek.EventType('FL')


@pytest.mark.remote_data
def test_ssw_latest_events_flares(flare_search):
    result = Fido.search(*flare_search, attrs.hek.FRM.Name == 'SSW Latest Events')
    assert len(result[0]) == 2


@pytest.mark.remote_data
def test_not_ssw_latest_events_flares(flare_search):
    result = Fido.search(*flare_search, attrs.hek.FRM.Name != 'SSW Latest Events')
    assert len(result[0]) == 19


@pytest.mark.remote_data
def test_flares_peak_flux(flare_search):
    result = Fido.search(*flare_search, attrs.hek.FL.PeakFlux > 4000.0)
    assert len(result[0]) == 1
    assert result[0]['fl_peakflux'] > 4000.0*u.DN/(u.pix*u.s)


@pytest.mark.remote_data
def test_flares_peak_flux_and_position(flare_search):
    result = Fido.search(*flare_search,
                         attrs.hek.Event.Coord1 > 800,
                         attrs.hek.FL.PeakFlux > 1000)
    assert len(result[0]) == 7
    assert all([c.Tx > 800*u.arcsec for c in result[0]['event_coord']])
    assert (result[0]['fl_peakflux'] > 1000.0*u.DN/(u.pix*u.s)).all()


@pytest.mark.remote_data
def test_flares_python_logical_ops(flare_search):
    result = Fido.search(*flare_search,
                         (attrs.hek.Event.Coord1 > 50) and (attrs.hek.FL.PeakFlux > 1000))
    assert len(result[0]) == 7
    assert all([c.Tx > 50*u.arcsec for c in result[0]['event_coord']])
    assert (result[0]['fl_peakflux'] > 1000.0*u.DN/(u.pix*u.s)).all()


@pytest.mark.remote_data
@pytest.mark.parametrize("event_type", [
    'AR',
    'CE',
    'CD',
    'CW',
    'FI',
    'FE',
    'FA',
    'LP',
    'OS',
    'SS',
    'EF',
    'CJ',
    'PG',
    'OT',
    'NR',
    'SG',
    'SP',
    'CR',
    'CC',
    'ER',
    'TO',
    'HY',
    'BU',
    'EE',
    'PB',
    'PT',
])
def test_event_types(event_type):
    # Smoke test for all event types
    _ = Fido.search(attrs.Time('2017/09/06 11:59:04', '2017/09/06 17:05:04'), attrs.hek.EventType(event_type))


@pytest.mark.remote_data
def test_raw_hek_result_preserved(hek_result):
    assert hasattr(hek_result, 'raw')
    assert isinstance(hek_result.raw, Table)
    # Check that separate event_coord columns are still present
    removed_columns = ['event_coord1', 'event_coord2', 'event_coord3', 'event_coordsys']
    for col in removed_columns:
        assert col in hek_result.raw.colnames
    # Check that times are still strings
    assert np.issubdtype(hek_result.raw['event_starttime'].dtype, np.str_)
    assert np.issubdtype(hek_result.raw['event_endtime'].dtype, np.str_)
    # Check that chaincodes are still strings
    for coord_attr in _get_coord_attributes():
        if not coord_attr.get('is_chaincode', False):
            continue
        assert np.issubdtype(hek_result.raw[coord_attr['name']].dtype, np.dtype('U'))
    # Check that quantities are still floats and that unit columns still exist
    for unit_attr in _get_unit_attributes():
        if (name := unit_attr['name']) not in hek_result.raw.colnames:
            continue
        column_dtype = hek_result.raw[name].dtype
        if unit_attr.get('unit_prop') is not None:
            # NOTE: Columns have object dtype if there are Nones present
            assert np.issubdtype(column_dtype, np.float64) | np.issubdtype(column_dtype, np.object_)
        elif unit_attr.get('is_unit_prop', False):
            assert np.issubdtype(column_dtype, np.str_)
